""" Contains the base class for Playgrounds.

Playground class should be inherited to create environments
where agents can play.
Playground defines the physics and mechanics of the game, and manages
how elements interact with each other.

Examples can be found in :
    - spg/playgrounds/empty.py
    - spg/playgrounds/collection
"""
# pylint: disable=too-many-public-methods

from __future__ import annotations

from typing import Dict, List, Optional, Tuple, Type, Union

import matplotlib.pyplot as plt
import numpy as np
import pymunk
import pymunk.matplotlib_util
from arcade import Window

from spg.element.element import PhysicalElement, SceneElement

from ..agent import Agent
from ..agent.communicator import Communicator, Message
from ..agent.controller import Command, Controller
from ..agent.part import AnchoredPart, PhysicalPart
from ..agent.sensor import RayCompute, RaySensor, Sensor, SensorValue
from ..entity import EmbodiedEntity, Entity, InteractiveAnchored
from ..utils.definitions import (
    PYMUNK_STEPS,
    SPACE_DAMPING,
    CollisionTypes,
    PymunkCollisionCategories,
)
from ..utils.position import Coordinate
from .collision_handlers import disabler_disables_device, grasper_grasps_graspable

# pylint: disable=unused-argument
# pylint: disable=line-too-long

CommandsDict = Dict[Agent, Dict[Controller, Command]]
TargetCommunicator = Optional[Union[Communicator, List[Communicator]]]
SentMessagesDict = Dict[Agent, Dict[Communicator, Tuple[TargetCommunicator, Message]]]

ObservationsDict = Dict[Agent, Dict[Sensor, SensorValue]]
ReceivedMessagesDict = Dict[Agent, Dict[Communicator, Tuple[Communicator, Message]]]
RewardsDict = Dict[Agent, float]


class Playground:
    """Playground is a Base Class that manages the physical simulation.

    Playground manages the interactions between Agents and Scene Elements.

    Attributes:
        initial_agent_coordinates: position or PositionAreaSampler,
            Starting position of an agent (single agent).
        done: bool, True if the playground reached termination.

    Notes:
          In the case of multi-agent setting,
          individual initial positions can be defined when
          instantiating the playground.

          Always reset the playground before starting a run.

    """

    # pylint: disable=too-many-instance-attributes

    time_limit = None
    time_limit_reached_reward = None

    def __init__(
        self,
        size: Optional[Tuple[int, int]] = None,
        seed: Optional[int] = None,
        background: Optional[
            Union[Tuple[int, int, int], List[int], Tuple[int, int, int, int]]
        ] = None,
        use_shaders=True,
    ):

        # Random number generator for replication, rewind, etc.
        self._rng = np.random.default_rng(seed)

        # By default, size is infinite and center is at (0,0)
        self._center = (0, 0)
        self._size = size
        self._width: Optional[int]
        self._height: Optional[int]

        if size:
            self._width, self._height = size
        else:
            self._width = self._height = None

        # Background color
        if not background:
            background = (0, 0, 0, 0)

        self._background = background

        # Initialization of the pymunk space, modelling all the physics
        self._space = self._initialize_space()

        # Lists containing elements in the playground
        self._elements: List[SceneElement] = []
        self._agents: List[Agent] = []
        self._teams = {}

        # Private attributes for managing interactions in playground
        self._done: bool = False
        self._timestep: int = 0

        # Mappings
        self._shapes_to_entities: Dict[pymunk.Shape, EmbodiedEntity] = {}
        self._name_to_agents: Dict[str, Agent] = {}
        self._uids_to_entities: Dict[int, Entity] = {}

        self._handle_interactions()
        self._views = []

        # Arcade window necessary to create contexts, views, sensors and gui
        self._window = Window(1, 1, visible=False, antialiasing=True)  # type: ignore
        self._window.ctx.blend_func = self._window.ctx.ONE, self._window.ctx.ZERO

        self._ray_compute = None
        self._use_shaders = use_shaders

    def debug_draw(self, plt_width=10, center=None, size=None):

        # todo: replace magic number by size calculation
        # depending on object position (boundingbox)

        if not center:
            center = self._center

        if not size:
            if not self._size:
                raise ValueError("Size must be set for display")
            size = self._size

        fig_size = (plt_width, plt_width / size[0] * size[1])

        fig = plt.figure(figsize=fig_size)

        ax = plt.axes(
            xlim=(center[0] - size[0] / 2, center[0] + size[0] / 2),
            ylim=(center[1] - size[1] / 2, center[1] + size[1] / 2),
        )
        ax.set_aspect("equal")

        options = pymunk.matplotlib_util.DrawOptions(ax)
        options.collision_point_color = (10, 20, 30, 40)
        self._space.debug_draw(options)
        # ax.invert_yaxis()
        plt.show()
        del fig

    @property
    def window(self):
        return self._window

    @property
    def ray_compute(self):

        if not self._ray_compute:
            assert self._size
            self._ray_compute = RayCompute(
                self, self._size, self._center, zoom=1, use_shader=self._use_shaders
            )

        return self._ray_compute

    @property
    def initial_agent_coordinates(self):
        return (0, 0), 0

    @property
    def background(self):
        return self._background

    @property
    def rng(self):
        return self._rng

    @property
    def size(self):
        return self._size

    @property
    def center(self):
        return self._center

    @property
    def timestep(self):
        return self._timestep

    @property
    def done(self):
        return self._done

    #################
    # Pymunk space
    #################

    @property
    def space(self):
        return self._space

    @staticmethod
    def _initialize_space() -> pymunk.Space:
        """Method to initialize Pymunk empty space for 2D physics.

        Returns: Pymunk Space

        """
        space = pymunk.Space()
        space.gravity = pymunk.Vec2d(0.0, 0.0)
        space.damping = SPACE_DAMPING

        return space

    ###############
    # Entities
    ###############

    def _get_identifier(self, entity: Entity):

        uid = None
        background = (
            self._background[0]
            + self._background[1] * 256
            + self._background[2] * 256 * 256
        )

        while True:
            a = self._rng.integers(0, 2**24)

            if a not in self._uids_to_entities and a != background:
                uid = a
                break

        name = entity.name
        if not name:
            name = type(entity).__name__ + "_" + str(uid)

        if name in [ent.name for ent in self.elements]:
            raise ValueError("Entity with this name already in Playground")

        return uid, name

    def get_entity_from_uid(self, uid):

        return self._uids_to_entities[uid]

    @property
    def agents(self):
        return [agent for agent in self._agents if not agent.removed]

    @property
    def elements(self):
        return [elem for elem in self._elements if not elem.removed]

    ###########
    # TEAMS
    ###########

    @property
    def teams(self):
        return self._teams

    def add_team(self, team):

        team_index = len(PymunkCollisionCategories) + len(self._teams) + 1
        self._teams[team] = team_index

    def _update_all_team_filters(self):

        for element in self._elements:
            element.update_team_filter()

        for agent in self._agents:
            agent.update_team_filter()

    ###############
    # STEP
    ###############

    def step(
        self,
        commands: Optional[CommandsDict] = None,
        messages: Optional[SentMessagesDict] = None,
        pymunk_steps: int = PYMUNK_STEPS,
    ):
        """Update the Playground

        Updates the Playground.
        Time moves by one unit of time.

        Args:
            pymunk_steps: Number of steps for the pymunk physics engine to run.

        Notes:
            pymunk_steps only influences the micro-steps that
            the physical engine is taking to render
            the movement of objects and collisions.
            From an external point of view,
            one unit of time passes independent on the number
            of pymunk_steps.

        """

        obs, mess, rew = None, None, None

        self._pre_step()

        if not self._done:

            self._apply_commands(commands)

            for _ in range(pymunk_steps):
                self.space.step(1.0 / pymunk_steps)

            self._post_step()
            self._done = self._has_terminated()

            obs = self._compute_observations()
            rew = {agent: agent.reward for agent in self._agents}
            if messages:
                mess = self._transmit_messages(messages)

            self._timestep += 1

        return obs, mess, rew, self._done

    def _pre_step(self):

        for element in self.elements:
            element.pre_step()

        for agent in self.agents:
            agent.pre_step()

    def _post_step(self):

        for element in self.elements:
            element.post_step()

        for agent in self.agents:
            agent.post_step()

    def _apply_commands(self, commands):

        if not commands:
            return

        for agent, command_dict in commands.items():
            agent.receive_commands(command_dict)

        for agent in self._agents:
            agent.apply_commands()

    def _transmit_messages(self, messages):

        msgs = {agent: {} for agent in self.agents}

        all_sent_messages = [
            (agent, comm_source, target)
            for agent, comms_dict in messages.items()
            for comm_source, target in comms_dict.items()
        ]

        for agent, comm_source, target in all_sent_messages:

            assert comm_source.agent is agent

            comm_target, message = target
            msg = comm_source.send(message)

            if not msg:
                continue

            if isinstance(comm_target, list):

                for targ in comm_target:
                    assert isinstance(targ, Communicator)
                    received_msg = targ.receive(comm_source, msg)

                    if received_msg:
                        msgs[targ.agent][targ] = (comm_source, received_msg)

            elif isinstance(comm_target, Communicator):
                received_msg = comm_target.receive(comm_source, msg)

                if received_msg:
                    msgs[comm_target.agent][comm_target] = (comm_source, received_msg)

            elif comm_target is None:
                for agent in self.agents:
                    for comm in agent.communicators:
                        received_msg = comm.receive(comm_source, msg)
                        if received_msg:
                            msgs[comm.agent][comm] = (comm_source, received_msg)

            else:
                raise ValueError

        return msgs

    def _compute_observations(self):

        if self._ray_compute:
            self._ray_compute.update_sensors()

        obs = {}
        for agent in self.agents:
            obs[agent] = agent.compute_observations()

        return obs

    def _has_terminated(self):
        return False

    def reset(self):
        """
        Reset the Playground to its initial state.
        """
        # reset elements that are still in playground
        for element in self._elements:

            if element.temporary:
                self.remove(element, definitive=True)
            elif element.removed:
                self.add(element, from_removed=True)
            elif isinstance(element, PhysicalElement) and element.movable:
                assert element.initial_coordinates
                element.move_to(
                    element.initial_coordinates,
                    element.allow_overlapping,
                )
            elif isinstance(element, EmbodiedEntity) and element.moved:
                assert element.initial_coordinates
                element.move_to(
                    element.initial_coordinates,
                    element.allow_overlapping,
                )

            element.reset()

        for agent in self._agents:
            agent.reset()
            if agent.removed:
                self.add(agent, from_removed=True)
            else:
                agent.base.move_to(
                    agent.initial_coordinates,
                    allow_overlapping=agent.allow_overlapping,
                    move_anchors=True,
                )

        for view in self._views:
            view.update(force=True)

        self._timestep = 0
        self._done = False

        obs = self._compute_observations()

        return obs, None, None, self._done

    # ADD REMOVE ENTITIES

    def add(
        self,
        entity,
        initial_coordinates=None,
        allow_overlapping=True,
        from_removed=False,
    ):

        entity.playground = self

        if from_removed:
            initial_coordinates = entity.initial_coordinates
            allow_overlapping = entity.allow_overlapping

        self._add_to_space(entity, initial_coordinates, allow_overlapping)

        if not from_removed:
            self._add_to_mappings(entity)

        self._add_to_views(entity)

        if isinstance(entity, Agent):
            self.add(
                entity.base,
                initial_coordinates=initial_coordinates,
                allow_overlapping=allow_overlapping,
                from_removed=from_removed,
            )

        if isinstance(entity, PhysicalPart):

            for part in entity.anchored:
                self.add(
                    part, allow_overlapping=allow_overlapping, from_removed=from_removed
                )

            for device in entity.devices:
                self.add(
                    device,
                    allow_overlapping=allow_overlapping,
                    from_removed=from_removed,
                )

                if isinstance(device, RaySensor):
                    self.ray_compute.add(device)

        elif isinstance(entity, PhysicalElement):
            for interactive in entity.interactives:
                self.add(
                    interactive,
                    allow_overlapping=allow_overlapping,
                    from_removed=from_removed,
                )

        self._update_teams(entity)

    def _add_to_space(self, entity, initial_coordinates, allow_overlapping):

        entity.removed = False

        if isinstance(entity, InteractiveAnchored):
            self._space.add(*entity.pm_shapes)
            return

        if isinstance(entity, AnchoredPart):
            initial_coordinates = entity.get_init_coordinates()

        elif isinstance(entity, Agent) and not initial_coordinates:
            initial_coordinates = self.initial_agent_coordinates

        elif entity.initial_coordinates:
            initial_coordinates = entity.initial_coordinates

        if not initial_coordinates:
            raise ValueError(
                "Either initial coordinate or size of the environment should be set"
            )

        entity.initial_coordinates = initial_coordinates
        entity.allow_overlapping = allow_overlapping

        if isinstance(entity, Agent):
            return

        self._space.add(*entity.pm_elements)
        entity.move_to(
            initial_coordinates,
            allow_overlapping=allow_overlapping,
        )

        if isinstance(entity, AnchoredPart):
            entity.attach_to_anchor()
            self._space.add(*entity.pm_joints)

    def _add_to_mappings(self, entity):

        entity.uid, entity.name = self._get_identifier(entity)

        self._uids_to_entities[entity.uid] = entity

        if isinstance(entity, Agent):
            self._agents.append(entity)
            self._name_to_agents[entity.name] = entity

        elif isinstance(entity, SceneElement):
            self._elements.append(entity)

        if isinstance(entity, EmbodiedEntity):
            for pm_shape in entity.pm_shapes:
                self._shapes_to_entities[pm_shape] = entity

    def _add_to_views(self, entity):

        if isinstance(entity, Agent):
            return

        for view in self._views:
            view.add(entity)

    def _update_teams(self, entity):

        if not isinstance(entity, (SceneElement, Agent)):
            return

        new_team = False
        for team in entity.teams:
            if team not in self._teams:
                self.add_team(team)
                new_team = True

        if new_team:
            self._update_all_team_filters()

        entity.update_team_filter()

    def remove(self, entity, definitive=False):

        self._remove_from_space(entity)
        self._remove_from_views(entity)

        if definitive:
            self._remove_from_mappings(entity)

        if isinstance(entity, Agent):
            self.remove(entity.base, definitive)

        if isinstance(entity, PhysicalPart):
            for part in entity.anchored:
                self.remove(part, definitive)

            for device in entity.devices:
                self.remove(device, definitive)

        elif isinstance(entity, PhysicalElement):

            for interactive in entity.interactives:
                self.remove(interactive, definitive)

            for grasper in entity.grasped_by:
                grasper.release(entity)

        entity.removed = True

    def _remove_from_space(self, entity):

        if isinstance(entity, Agent):
            return

        self._space.remove(*entity.pm_elements)

        if isinstance(entity, AnchoredPart):
            self._space.remove(*entity.pm_joints)

    def _remove_from_mappings(self, entity):

        assert entity.uid

        self._uids_to_entities.pop(entity.uid)

        if isinstance(entity, Agent):
            self._agents.remove(entity)

            assert entity.name
            self._name_to_agents.pop(entity.name)

        elif isinstance(entity, SceneElement):
            self._elements.remove(entity)

        if not isinstance(entity, Agent):
            for pm_shape in entity.pm_shapes:
                self._shapes_to_entities.pop(pm_shape)

        entity.playground = None

    def _remove_from_views(self, entity):

        if isinstance(entity, Agent):
            return
        for view in self._views:
            view.remove(entity)

    def add_view(self, view):

        for entity in self.elements:
            view.add(entity)

            if isinstance(entity, PhysicalElement):
                for interactive in entity.interactives:
                    view.add(interactive)

        for agent in self.agents:
            for part in agent.parts:
                view.add(part)

                for device in part.devices:
                    view.add(device)

        self._views.append(view)

    def within_playground(
        self,
        entity: Optional[Union[Agent, EmbodiedEntity]] = None,
        coordinates: Optional[Coordinate] = None,
    ):

        if not self._size:
            return True

        if isinstance(entity, Agent):
            for part in entity.parts:
                if not self.within_playground(part):
                    return False

        if entity:
            position = entity.position
        elif coordinates:
            position = coordinates[0]
        else:
            raise ValueError("entity or coordinates must be specified")

        if not -self._size[0] / 2 <= position[0] <= self._size[0] / 2:
            return False

        if not -self._size[1] / 2 <= position[1] <= self._size[1] / 2:
            return False

        return True

    def overlaps(self, entity: EmbodiedEntity, coordinates):
        """Tests whether new coordinate would lead to physical collision"""

        dummy_body = pymunk.Body(body_type=pymunk.Body.STATIC)
        dummy_shapes = []
        for pm_shape in entity.pm_shapes:
            dummy_shape = pm_shape.copy()
            dummy_shape.body = dummy_body
            dummy_shape.sensor = True
            dummy_shapes.append(dummy_shape)

        self.space.add(dummy_body, *dummy_shapes)

        dummy_body.position, dummy_body.angle = coordinates
        self.space.reindex_static()

        overlaps = []
        for dummy_shape in dummy_shapes:
            overlaps += self.space.shape_query(dummy_shape)
        self.space.remove(dummy_body, *dummy_shapes)

        # remove sensor shapes
        overlaps = [
            elem
            for elem in overlaps
            if elem.shape
            and not elem.shape.sensor
            and elem.shape not in entity.pm_shapes
        ]

        if isinstance(entity, PhysicalPart):
            agent_shapes = []
            for part in entity.agent.parts:
                agent_shapes += part.pm_shapes

            overlaps = [elem for elem in overlaps if elem.shape not in agent_shapes]

        self.space.reindex_static()

        return bool(overlaps)

    def get_closest_agent(self, entity: EmbodiedEntity) -> Agent:
        return min(self.agents, key=lambda a: entity.position.get_dist_sqrd(a.position))

    def get_entity_from_shape(self, shape: pymunk.Shape):
        assert shape in self._shapes_to_entities

        entity = self._shapes_to_entities[shape]

        return entity

    def _handle_interactions(self):

        # Order is important

        self.add_interaction(
            CollisionTypes.GRASPER, CollisionTypes.GRASPABLE, grasper_grasps_graspable
        )

        self.add_interaction(
            CollisionTypes.DISABLER, CollisionTypes.DEVICE, disabler_disables_device
        )
        # self.add_interaction(CollisionTypes.PART, CollisionTypes.CONTACT,

    #                              agent_touches_element)
    #         self.add_interaction(CollisionTypes.PART, CollisionTypes.ACTIVABLE,
    #                              agent_activates_element)
    #         self.add_interaction(CollisionTypes.GEM,
    #                              CollisionTypes.ACTIVABLE_BY_GEM,
    #                              gem_activates_element)
    #         self.add_interaction(CollisionTypes.PART, CollisionTypes.TELEPORT,
    #                              agent_teleports)

    def add_interaction(
        self,
        collision_type_1: CollisionTypes,
        collision_type_2: CollisionTypes,
        interaction_function,
    ):
        """

        Args:
            collision_type_1: collision type of the first entity
            collision_type_2: collision type of the second entity
            interaction_function: function that handles the interaction

        Returns: None

        """

        handler = self.space.add_collision_handler(collision_type_1, collision_type_2)
        handler.pre_solve = interaction_function
        handler.data["playground"] = self

    # def __del__(self):
    #     self.close()
    #     # gc.collect()
    #     super().__del__()


class PlaygroundRegister:
    """
    Class to register Playgrounds.
    """

    playgrounds: Dict[str, Dict[str, Type[Playground]]] = {}

    @classmethod
    def register(
        cls,
        playground_group: str,
        playground_name: str,
    ):
        """
        Registers a playground
        """

        def decorator(subclass):

            if playground_group not in cls.playgrounds:
                cls.playgrounds[playground_group] = {}

            if playground_name in cls.playgrounds[playground_group]:
                raise ValueError(playground_name + " already registered")

            cls.playgrounds[playground_group][playground_name] = subclass
            return subclass

        return decorator



from abc import ABC

from .sensor import ExternalSensor


class TopDownSensor(ExternalSensor, ABC):

    """
    Base class for Image Based sensors.
    Image based sensors are computed using the top-down rendering of the playground.

    """



from typing import List, Optional, Type, Union

from ..agent.communicator import Communicator
from ..agent.controller import Controller
from ..agent.device import Device
from ..agent.sensor import Sensor
from ..utils.definitions import CollisionTypes
from .element import ZoneElement


class Disabler(ZoneElement):
    def __init__(self, disable_cls: Union[List[Type[Device]], Type[Device]]):

        super().__init__(filename=":spg:platformer/tiles/area_forbidden.png")

        self._disable_cls = disable_cls

    @property
    def _collision_type(self):
        return CollisionTypes.DISABLER

    def disable(self, device: Device):

        if isinstance(device, self._disable_cls):
            device.disable()


class SensorDisabler(Disabler):
    def __init__(
        self, sensors: Optional[Union[List[Type[Sensor]], Type[Sensor]]] = None
    ):

        if not sensors:
            sensors = Sensor

        super().__init__(disable_cls=sensors)


class ControllerDisabler(Disabler):
    def __init__(
        self,
        controllers: Optional[Union[List[Type[Controller]], Type[Controller]]] = None,
    ):

        if not controllers:
            controllers = Controller

        super().__init__(disable_cls=controllers)


class CommunicatorDisabler(Disabler):
    def __init__(
        self,
        communicators: Optional[
            Union[List[Type[Communicator]], Type[Communicator]]
        ] = None,
    ):

        if not communicators:
            communicators = Communicator

        super().__init__(disable_cls=communicators)



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from ..device import Device

if TYPE_CHECKING:
    from ...utils.position import Coordinate


INTERACTOR_RADIUS = 5


class Interactor(Device):
    def __init__(self, anchor, **kwargs):

        super().__init__(anchor=anchor, **kwargs)

        # Point on the anchor
        self._anchor_coordinates = ((0, 0), 0)

    @property
    def anchor_coordinates(self):
        return self._anchor_coordinates

    @anchor_coordinates.setter
    def anchor_coordinates(self, coord: Coordinate):
        self._anchor_coordinates = coord

    def relative_position(self):

        return (self.position - self._anchor.position).rotated(-self._anchor.angle)

    def relative_angle(self):
        return self.angle - self._anchor.angle


class ActiveInteractor(Interactor):
    @abstractmethod
    def apply_commands(self, **kwargs):
        ...



from __future__ import annotations

import numpy as np
import pymunk

from ...utils.definitions import ANGULAR_VELOCITY, LINEAR_FORCE
from ..controller import CenteredContinuousController
from .part import AnchoredPart, PhysicalPart


class ForwardBase(PhysicalPart):
    def __init__(
        self,
        linear_ratio: float = 1,
        angular_ratio: float = 1,
        **kwargs,
    ):

        super().__init__(
            mass=30,
            filename=":spg:puzzle/element/element_blue_square.png",
            sprite_front_is_up=True,
            shape_approximation="decomposition",
            **kwargs,
        )

        self.forward_controller = CenteredContinuousController("forward")
        self.add(self.forward_controller)

        self.angular_vel_controller = CenteredContinuousController("angular")
        self.add(self.angular_vel_controller)

        self.linear_ratio = LINEAR_FORCE * linear_ratio
        self.angular_ratio = ANGULAR_VELOCITY * angular_ratio

    def _apply_commands(self, **kwargs):

        command_value = self.forward_controller.command_value

        self._pm_body.apply_force_at_local_point(
            pymunk.Vec2d(command_value, 0) * self.linear_ratio, (0, 0)
        )

        command_value = self.angular_vel_controller.command_value
        self._pm_body.angular_velocity = command_value * self.angular_ratio


class Head(AnchoredPart):
    def __init__(self, **kwargs):
        super().__init__(
            mass=10,
            radius=10,
            filename=":spg:puzzle/element/element_blue_polygon.png",
            sprite_front_is_up=True,
            **kwargs,
        )
        self.joint_controller = CenteredContinuousController("head")
        self.add(self.joint_controller)

    def _apply_commands(self, **kwargs):

        assert self._anchor
        assert self._motor
        angle_offset = self._anchor_coordinates[1]

        value = self.joint_controller.command_value

        theta_part = self.angle
        theta_anchor = self._anchor.angle

        angle_centered = theta_part - (theta_anchor + angle_offset)
        angle_centered = angle_centered % (2 * np.pi)
        angle_centered = (
            angle_centered - 2 * np.pi if angle_centered > np.pi else angle_centered
        )

        # Do not set the motor if the limb is close to limit
        if (angle_centered < -self._rotation_range / 2 + np.pi / 20) and value < 0:
            self._motor.rate = 0

        elif (angle_centered > self._rotation_range / 2 - np.pi / 20) and value > 0:
            self._motor.rate = 0

        else:
            self._motor.rate = -value * ANGULAR_VELOCITY

    @property
    def _pivot_position(self):
        return (0, 0)



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Optional, Tuple

import arcade

from .view import TopDownView

if TYPE_CHECKING:
    from ..agent import Agent
    from ..agent.controller import Command, Controller
    from ..playground import Playground


class GUI(TopDownView):
    def __init__(
        self,
        playground: Playground,
        keyboard_agent: Optional[Agent] = None,
        size: Optional[Tuple[int, int]] = None,
        center: Tuple[float, float] = (0, 0),
        zoom: float = 1,
        display_uid: bool = False,
        draw_transparent: bool = True,
        draw_interactive: bool = True,
        draw_zone: bool = True,
        draw_sensors: bool = False,
        print_rewards: bool = True,
        print_messages: bool = True,
        random_agents: bool = True,
    ) -> None:
        super().__init__(
            playground,
            size,
            center,
            zoom,
            display_uid,
            draw_transparent,
            draw_interactive,
            draw_zone,
        )

        self._playground.window.set_size(*self._size)
        self._playground.window.set_visible(True)

        self._keyboard_agent = keyboard_agent
        self._random_agents = random_agents

        self._agent_commands: Dict[Controller, Command] = {}
        self._message = None

        self.print_rewards = print_rewards
        self.print_messages = print_messages

        self._playground.window.on_draw = self.on_draw
        self._playground.window.on_update = self.on_update
        self._playground.window.on_key_press = self.on_key_press
        self._playground.window.on_key_release = self.on_key_release

        self._draw_sensors = draw_sensors

    def run(self):
        self._playground.window.run()

    def on_draw(self):

        self._playground.window.clear()
        self._fbo.use()
        self.update()

    def on_update(self, _):

        commands = self._get_commands()

        self._playground.step(commands=commands, messages=self._message)

        if self.print_rewards:

            for agent in self._playground.agents:
                if agent.reward != 0:
                    print(agent.reward)

        if self.print_messages:

            for agent in self._playground.agents:
                for comm in agent.communicators:
                    for _, msg in comm.received_messages:
                        print(f"Agent {agent.name} received message {msg}")

        self._message = {}

    def _get_commands(self):

        command_dict = {}

        if self._random_agents:
            for agent in self._playground.agents:
                command_dict[agent] = agent.get_random_commands()

        if self._keyboard_agent:
            command_dict[self._keyboard_agent] = self._agent_commands

        return command_dict

    def update(self, force=False):

        self.update_sprites(force)

        self._playground.window.use()

        self._playground.window.clear(self._background)

        if self._draw_sensors:

            for agent in self._playground.agents:
                for sensor in agent.sensors:
                    sensor.draw()

        # Change projection to match the contents

        self._transparent_sprites.draw(pixelated=True)
        self._interactive_sprites.draw(pixelated=True)
        self._zone_sprites.draw(pixelated=True)
        self._visible_sprites.draw(pixelated=True)
        self._traversable_sprites.draw(pixelated=True)

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed."""

        self._keyboard_agent_key_press(key, modifiers)

        if key == arcade.key.Q:
            self._playground.window.close()

        if key == arcade.key.R:
            self._playground.reset()

        if key == arcade.key.S:
            self._draw_sensors = not self._draw_sensors

    def on_key_release(self, key, modifiers):

        self._keyboard_agent_key_release(key, modifiers)

    def _keyboard_agent_key_press(self, key, modifiers):
        """
        Implement for your agent
        """

    def _keyboard_agent_key_release(self, key, modifiers):
        """
        Implement for your agent
        """


class HeadAgentGUI(GUI):
    def _keyboard_agent_key_press(self, key, modifiers):

        if key == arcade.key.UP:
            self._agent_commands["forward"] = 0.2
        elif key == arcade.key.DOWN:
            self._agent_commands["forward"] = -0.2

        if not modifiers & arcade.key.MOD_SHIFT:
            if key == arcade.key.LEFT:
                self._agent_commands["angular"] = 0.2
            elif key == arcade.key.RIGHT:
                self._agent_commands["angular"] = -0.2
        else:
            if key == arcade.key.LEFT:
                self._agent_commands["head"] = 0.1
            elif key == arcade.key.RIGHT:
                self._agent_commands["head"] = -0.1

        if key == arcade.key.M:
            self._message = {
                self._keyboard_agent: {
                    self._keyboard_agent.comm: (
                        None,
                        f"Currently at timestep {self._playground.timestep}",
                    )
                }
            }
            print(f"Agent {self._keyboard_agent.name} sends message")

        if key == arcade.key.G:
            self._agent_commands["grasper"] = 1

    def _keyboard_agent_key_release(self, key, modifiers):

        if key == arcade.key.UP:
            self._agent_commands["forward"] = 0
        elif key == arcade.key.DOWN:
            self._agent_commands["forward"] = 0
        if not modifiers & arcade.key.MOD_SHIFT:
            if key == arcade.key.LEFT:
                self._agent_commands["angular"] = 0
            elif key == arcade.key.RIGHT:

                self._agent_commands["angular"] = 0
        else:
            if key == arcade.key.LEFT:
                self._agent_commands["head"] = 0
            elif key == arcade.key.RIGHT:
                self._agent_commands["head"] = 0

        if key == arcade.key.G:
            self._agent_commands["grasper"] = 0



from ..utils.definitions import CollisionTypes
from .element import PhysicalElement, RewardElement


class Chest(PhysicalElement):
    def __init__(self, color=None):

        super().__init__(
            radius=15,
            filename=":spg:spg/goal.png",
            color=color,
        )

    def _set_pm_collision_type(self):
        for pm_shape in self._pm_shapes:
            pm_shape.collision_type = CollisionTypes.ACTIVABLE_BY_GEM

    def activate(self, entity: RewardElement):

        assert self._playground

        if entity.graspable:
            for grasper in entity.grasped_by:
                agent = grasper.anchor.agent
                agent.reward += entity.reward / len(entity.grasped_by)

        else:
            agent = self._playground.get_closest_agent(self)
            agent.reward += entity.reward

        self._playground.remove(entity)



# Checkpoints for easy rewind
#       self._checkpoints: Dict[Timestep, bytes] = {}
#       self._actions: Dict[Timestep, Tuple[AgentCommandsDict, AgentMessagesDict]] = {}

# Make checkpoints a decorator?


#  ################
#   # Checkpoints
#   ################

#   def save_checkpoint(self):
#       self._checkpoints[self._timestep] = self._get_checkpoint()

#   def delete_checkpoints(self):
#       self._checkpoints = {}

#   def _get_checkpoint(self):
#       pg = copy.copy(self)
#       pg._checkpoints = {}
#       return pickle.dumps(pg)

#   def _save_actions_for_replay(self, commands, messages):
#       self._actions[self._timestep] = commands, messages


#   def rewind(self, timesteps_rewind, random_alternate: bool = False, **kwargs):

#       # load pg
#       valid_checkpoints = [ts for ts in self._checkpoints.keys()
# if ts <= self._timestep - timesteps_rewind]
#       if not valid_checkpoints:
#           raise ValueError('no checkpoints were saved for rewind')

#       ts_checkpoint = max(valid_checkpoints)
#       pg = pickle.loads(self._checkpoints[ts_checkpoint])

#       # Save before overriding playground
#       future_actions = [self._actions[ts]
# for ts in range(ts_checkpoint, self._timestep-timesteps_rewind)]
#       rng = self._rng
#       checkpoints = self._checkpoints

#       # Time travel
#       self.__dict__.update(pg.__dict__)
#       if random_alternate:
#           self._rng = np.random.default_rng(seed=rng.integers(1000))

#       # Apply all action between checkpoint and rewind point
#       for act in future_actions:
#           self.step(actions=act, compute_observations=False)

#       self._checkpoints = {ts: checkpoint
# for ts, checkpoint in checkpoints.items() if ts <= self._timestep}
#       self._actions = {ts: action
# for ts, action in self._actions.items() if ts <= self._timestep}

#       for view in self._views:
#           view.reset()

#       obs = self._compute_observations(**kwargs)
#       rew = self._compute_rewards(**kwargs)

#       return obs, rew, self.done



from __future__ import annotations

from typing import Any, List, Optional, Tuple

from ..device import PocketDevice

Message = Any


COMM_COLOR = (123, 134, 13)


class Communicator(PocketDevice):
    def __init__(
        self,
        transmission_range: Optional[float] = None,
    ):
        """
        By default, Communicator has infinite range and infinite receiver capacity.
        However, it can only send one message at a time.


        Args:
            anchor:
            transmission_range:
            receiver_capacity:
        """

        super().__init__(color=COMM_COLOR)

        self._transmission_range = transmission_range

        self._comms_in_range: List[Communicator] = []
        self._received_messages: List[Tuple[Communicator, Message]] = []

    def pre_step(self):
        super().pre_step()
        self.update_list_comms_in_range()
        self._received_messages = []

    def reset(self):
        self.pre_step()
        super().reset()

    @property
    def transmission_range(self):
        return self._transmission_range

    def update_list_comms_in_range(self):

        assert self._playground

        comms = []
        for agent in self._playground.agents:
            comms += agent.communicators

        valid_comms = [com for com in comms if com is not self]

        for comm in valid_comms:
            if self.in_transmission_range(comm):
                self._comms_in_range.append(comm)

    @property
    def comms_in_range(self):
        return self._comms_in_range

    @property
    def received_messages(self):
        return self._received_messages

    def in_transmission_range(self, comm: Communicator):

        dist = comm.position.get_distance(self.position)

        # If both have infinite range:
        if not (comm.transmission_range or self._transmission_range):
            return True

        # If only one has infinite range:
        if (not comm.transmission_range) and self._transmission_range:
            if dist < self._transmission_range:
                return True

        elif comm.transmission_range and (not self._transmission_range):
            if dist < comm.transmission_range:
                return True

        elif dist < comm.transmission_range and dist < self._transmission_range:
            return True

        return False

    def send(
        self,
        msg: Message,
    ) -> Optional[Message]:

        if self._disabled:
            return None

        return msg

    def receive(self, sender, msg):

        if self._disabled or sender is self:
            return None

        if self.in_transmission_range(sender):
            self._received_messages.append((sender, msg))
            return msg

        return None


class LimitedCommunicator(Communicator):
    def __init__(self, capacity: Optional[int] = None, **kwargs):

        self._capacity = capacity

        super().__init__(**kwargs)

    @property
    def receiver_capacity(self):
        return self._capacity

    def receive(self, sender, msg):
        msg = super().receive(sender, msg)

        if self._capacity:
            self._received_messages.sort(
                key=(lambda s_m: self.position.get_distance(s_m[0].anchor.position))
            )
            self._received_messages = self._received_messages[: self._capacity]

        return msg



# from .layouts import SingleRoom, LineRooms, GridRooms
from .collision_handlers import get_colliding_entities
from .playground import Playground
from .room import ConnectedRooms, Room

__all__ = ["get_colliding_entities", "Playground", "Room", "ConnectedRooms"]



from .embodied import EmbodiedEntity
from .entity import Entity
from .interactive import (
    Graspable,
    InteractiveAnchored,
    InteractiveEntity,
    InteractiveZone,
)
from .physical import PhysicalEntity

__all__ = [
    "Entity",
    "EmbodiedEntity",
    "Graspable",
    "InteractiveZone",
    "InteractiveEntity",
    "InteractiveAnchored",
    "PhysicalEntity",
]



from __future__ import annotations

from abc import abstractmethod
from typing import List, Union

import numpy as np

from ..device import PocketDevice

Command = Union[float, int, bool]

CONTROLLER_COLOR = (123, 234, 213)


class Controller(PocketDevice):
    """
    Command classes define how parts can be controlled.
    It is used to control parts of an agent:
        - physical actions (movements)
        - interactive actions (eat, grasp, ...)
    """

    def __init__(self, name, hard_check=True, **_):

        super().__init__(name=name, color=CONTROLLER_COLOR)

        self._command = self.default
        self._hard_check = hard_check
        self._currently_disabled = False

    @property
    def _rng(self):
        if self._playground:
            return self._playground.rng

        return np.random.default_rng()

    @property
    @abstractmethod
    def default(self) -> Command:
        ...

    @abstractmethod
    def get_random_commands(self):
        ...

    @property
    def command(self):
        return self._command

    @command.setter
    def command(self, command):
        check_passed = self._check(command)

        if not check_passed and self._hard_check:
            raise ValueError(command)

        # Maybe replace by closest later?
        if not check_passed or self._currently_disabled:
            command = self.default

        self._command = command

    def pre_step(self):
        super().pre_step()
        self._command = self.default

    def post_step(self):
        self._currently_disabled = self._disabled

    def reset(self):
        self.pre_step()

    @abstractmethod
    def _check(self, command) -> bool:
        ...

    @property
    def command_value(self) -> Command:
        return self._command


####################
# DISCRETE COMMANDS
####################


class DiscreteController(Controller):
    """
    Discrete Commands.
    Command values can take a number within a list of integers.
    0 is always the default command, even if not given at initialization.
    """

    def __init__(
        self,
        name: str,
        command_values: List[int],
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)

        if not isinstance(command_values, (list, tuple)):
            raise ValueError("Set of values must be list or tuple")

        self._valid_command_values = command_values

    def _check(self, command) -> bool:
        return command in self._valid_command_values

    @property
    def default(self) -> int:
        return 0

    @property
    def valid_commands(self):
        return self._valid_command_values

    def get_random_commands(self):
        return self._playground.rng.choice(self._valid_command_values)


class BoolController(DiscreteController):
    def __init__(self, name, **kwargs):
        super().__init__(name=name, command_values=[0, 1], **kwargs)


class GrasperController(BoolController):
    pass


class RangeController(DiscreteController):
    def __init__(self, name, n: int, **kwargs):
        super().__init__(name=name, command_values=list(range(n)), **kwargs)


###################
# ContinuousCommand
###################


class ContinuousController(Controller):
    def __init__(self, name, min_value: float, max_value: float, **kwargs):

        super().__init__(name=name, **kwargs)

        if min_value > max_value:
            raise ValueError

        self._min = min_value
        self._max = max_value

    def _check(self, command) -> bool:

        return self._min <= command <= self._max

    @property
    def default(self) -> float:
        return 0

    @property
    def min(self) -> float:
        return self._min

    @property
    def max(self) -> float:
        return self._max

    def get_random_commands(self):
        return self._playground.rng.uniform(self._min, self._max)


class CenteredContinuousController(ContinuousController):
    def __init__(self, name, **kwargs):
        super().__init__(name=name, min_value=-1, max_value=1, **kwargs)


class NormalContinuousController(ContinuousController):
    def __init__(self, name, **kwargs):
        super().__init__(name=name, min_value=0, max_value=1, **kwargs)



from __future__ import annotations

from abc import ABC
from array import array
from os import path
from typing import TYPE_CHECKING, List

import arcade
import numpy as np

from ...utils.uid import id_to_pixel
from ...view import TopDownView
from .sensor import ExternalSensor

if TYPE_CHECKING:
    from ...playground import Playground


class RayCompute:
    def __init__(self, playground: Playground, size, center, zoom, use_shader=True):

        self._playground = playground
        self.ctx = playground.window.ctx

        self._use_shader = use_shader

        self._id_view = TopDownView(
            playground,
            size,
            center,
            zoom,
            display_uid=True,
            draw_interactive=False,
            draw_transparent=False,
            draw_zone=False,
        )
        self._color_view = TopDownView(
            playground,
            size,
            center,
            zoom,
            draw_interactive=False,
            draw_transparent=False,
            draw_zone=False,
        )

        self._sensors: List[RaySensor] = []

        if self._use_shader:

            self._view_params_buffer = self.ctx.buffer(
                data=array(
                    "f",
                    [
                        self._id_view.center[0],
                        self._id_view.center[1],
                        self._id_view.width,
                        self._id_view.height,
                        self._id_view.zoom,
                    ],
                )
            )

            self._view_params_buffer.bind_to_storage_buffer(binding=6)

            self._position_buffer = None
            self._param_buffer = None
            self._output_rays_buffer = None
            self._inv_buffer = None

            shader_dir = path.abspath(path.join(__file__, "../shaders"))

            with open(shader_dir + "/id_compute.glsl", "rt", encoding="utf-8") as f_id:
                self._source_compute_ids = f_id.read()

            with open(
                shader_dir + "/color_compute.glsl", "rt", encoding="utf-8"
            ) as f_col:
                self._source_compute_colors = f_col.read()

            self._id_shader = None
            self._color_shader = None

    @property
    def _n_sensors(self):
        return len(self._sensors)

    @property
    def _max_n_rays(self):
        return max(sensor.resolution for sensor in self._sensors)

    @property
    def _max_invisible(self):
        return 1 + max(len(sensor.invisible_ids) for sensor in self._sensors)

    def _generate_buffers(self):

        position_buffer = self.ctx.buffer(
            data=array("f", self._generate_position_buffer())
        )
        param_buffer = self.ctx.buffer(
            data=array("f", self._generate_parameter_buffer())
        )
        output_rays_buffer = self.ctx.buffer(
            data=array("f", self._generate_output_buffer())
        )
        inv_buffer = self.ctx.buffer(data=array("I", self._generate_invisible_buffer()))

        param_buffer.bind_to_storage_buffer(binding=2)
        position_buffer.bind_to_storage_buffer(binding=3)
        output_rays_buffer.bind_to_storage_buffer(binding=4)
        inv_buffer.bind_to_storage_buffer(binding=5)

        return position_buffer, param_buffer, output_rays_buffer, inv_buffer

    def _generate_parameter_buffer(self):

        for sensor in self._sensors:
            yield sensor.max_range
            yield sensor.fov
            yield sensor.resolution
            yield sensor.n_points

    def _generate_position_buffer(self):

        for sensor in self._sensors:
            yield sensor.position[0]
            yield sensor.position[1]
            yield sensor.angle

    def _generate_output_buffer(self):

        for _ in range(self._n_sensors):
            for _ in range(self._max_n_rays):

                # View Position
                yield 0.0
                yield 0.0

                # Abs Env Position
                yield 0.0
                yield 0.0

                # Rel Position
                yield 0.0
                yield 0.0

                # Sensor center on view
                yield 0.0
                yield 0.0

                # ID
                yield 0.0

                # Distance
                yield 0.0

                # Color
                yield 0.0
                yield 0.0
                yield 0.0

    def _generate_invisible_buffer(self):

        for sensor in self._sensors:

            count = 1
            yield sensor.anchor.uid

            for inv in sensor.invisible_ids:
                yield inv
                count += 1

            while count < self._max_invisible:
                yield 0
                count += 1

    def _generate_shaders(self):
        new_source = self._source_compute_ids
        new_source = new_source.replace("N_SENSORS", str(len(self._sensors)))
        new_source = new_source.replace("MAX_N_RAYS", str(self._max_n_rays))
        new_source = new_source.replace("MAX_N_INVISIBLE", str(self._max_invisible))
        id_shader = self.ctx.compute_shader(source=new_source)

        new_source = self._source_compute_colors
        new_source = new_source.replace("MAX_N_RAYS", str(self._max_n_rays))
        color_shader = self.ctx.compute_shader(source=new_source)

        return id_shader, color_shader

    def add(self, sensor):
        self._sensors.append(sensor)

        if self._use_shader:
            self._update_buffers_and_shaders()

    def _update_buffers_and_shaders(self):
        (
            self._position_buffer,
            self._param_buffer,
            self._output_rays_buffer,
            self._inv_buffer,
        ) = self._generate_buffers()

        self._id_shader, self._color_shader = self._generate_shaders()

    def update_sensors(self):

        if not self._sensors:
            return

        self._id_view.update(force=True)
        self._color_view.update(force=True)

        if self._use_shader:
            self._update_sensors_shaders()

        else:
            self._update_sensors_cpu()

    def _update_sensors_shaders(self):

        update_inv = False
        for sensor in self._sensors:
            if sensor.require_invisible_update:
                update_inv = True

        if update_inv:
            self._update_buffers_and_shaders()

        self._position_buffer = self.ctx.buffer(
            data=array("f", self._generate_position_buffer())
        )
        self._position_buffer.bind_to_storage_buffer(binding=3)

        self._id_view.texture.use()
        self._id_shader.run(group_x=self._n_sensors)

        self._color_view.texture.use()
        self._color_shader.run(group_x=self._n_sensors)

        hitpoints = np.frombuffer(
            self._output_rays_buffer.read(), dtype=np.float32
        ).reshape(self._n_sensors, self._max_n_rays, 13)

        for index, sensor in enumerate(self._sensors):
            sensor.update_hitpoints(hitpoints[index, : sensor.resolution, :])

    def _update_sensors_cpu(self):

        img_color = self._color_view.get_np_img()
        img_id = self._id_view.get_np_img()

        for sensor in self._sensors:

            end_positions = sensor.end_positions

            ray_start_x = (
                sensor.position[0] - self._id_view.center[0]
            ) * self._id_view.zoom + self._id_view.width / 2
            ray_start_y = (
                sensor.position[1] - self._id_view.center[1]
            ) * self._id_view.zoom + self._id_view.height / 2

            center_on_view = np.asarray((ray_start_x, ray_start_y)).reshape(2, 1)

            rays_end = center_on_view + end_positions * self._id_view.zoom

            points = np.linspace(center_on_view, rays_end, num=sensor.n_points)

            points[:, 0, :] = (points[:, 0, :] > 0) * points[:, 0, :]
            points[:, 0, :] = (points[:, 0, :] < self._id_view.width) * points[
                :, 0, :
            ] + (points[:, 0, :] >= self._id_view.width) * (self._id_view.width - 1)

            points[:, 1, :] = (points[:, 1, :] > 0) * points[:, 1, :]
            points[:, 1, :] = (points[:, 1, :] < self._id_view.height) * points[
                :, 1, :
            ] + (points[:, 1, :] >= self._id_view.height) * (self._id_view.height - 1)

            points = points.swapaxes(1, 2).reshape(-1, 2)

            rr, cc = points[:, 1].astype(np.int), points[:, 0].astype(np.int)

            # Get Ids
            pts = img_id[rr, cc]

            ids = 256 * 256 * pts[:, 2] + 256 * pts[:, 1] + pts[:, 0]

            #  remove invisible
            for inv in sensor.invisible_ids:
                ids = (ids != inv) * ids

            ids = ids.reshape(-1, sensor.resolution).transpose()

            # take index and value of first non-zero
            index_first_non_zero = np.argmax((ids != 0), axis=1)

            # Calculate hitpoints
            id_first_non_zero = ids[
                np.arange(len(index_first_non_zero)), index_first_non_zero
            ]

            points = points.reshape((-1, sensor.resolution, 2))
            view_position = points[
                index_first_non_zero, np.arange(len(index_first_non_zero)), :
            ]

            color = img_color[
                view_position[:, 1].astype(np.int), view_position[:, 0].astype(np.int)
            ]
            color[id_first_non_zero == 0] = (0, 0, 0)

            # Relative pos and distance
            rel_pos = center_on_view.transpose() - view_position
            distance = np.sqrt(rel_pos[:, 0] ** 2 + rel_pos[:, 1] ** 2)
            distance[id_first_non_zero == 0] = sensor.max_range
            distance = np.expand_dims(distance, -1)

            x_abs = (
                view_position[:, 0] - self._id_view.width / 2
            ) / self._id_view.zoom + self._id_view.center[0]

            y_abs = (
                view_position[:, 1] - self._id_view.height / 2
            ) / self._id_view.zoom + self._id_view.center[1]

            view_position[id_first_non_zero == 0, :] = rays_end[
                :, id_first_non_zero == 0
            ].transpose()
            abs_env_position = np.vstack((x_abs, y_abs)).transpose()

            center_on_view = np.broadcast_to(
                center_on_view.transpose(), (sensor.resolution, 2)
            )

            id_first_non_zero = np.expand_dims(id_first_non_zero, -1)

            hitpoints = np.hstack(
                (
                    view_position,
                    abs_env_position,
                    np.zeros((sensor.resolution, 2)),
                    center_on_view,
                    id_first_non_zero,
                    distance,
                    color,
                )
            )

            sensor.update_hitpoints(hitpoints)


class RaySensor(ExternalSensor, ABC):
    """
    Base class for Ray Based sensors.
    Ray sensors use Arcade shaders
    """

    def __init__(
        self,
        spatial_resolution: float = 1,
        **kwargs,
    ):

        super().__init__(**kwargs)

        self._spatial_resolution = spatial_resolution
        self._n_points = int(self._range / self._spatial_resolution)

        self._hitpoints = 0

    @property
    def spatial_resolution(self):
        return self._spatial_resolution

    @property
    def n_points(self):
        return self._n_points

    def update_hitpoints(self, hitpoints):
        self._hitpoints = hitpoints

    @property
    def end_positions(self):
        angles = np.array(
            [
                self.angle - self.fov / 2 + i_ray * self.fov / (self._resolution - 1)
                for i_ray in range(self._resolution)
            ]
        )
        x = self._range * np.cos(angles)
        y = self._range * np.sin(angles)
        return np.vstack((x, y))


class DistanceSensor(RaySensor):
    def _compute_raw_sensor(self):
        self._values = self._hitpoints[:, 9]

    def draw(self):

        if self._disabled:
            return
        view_xy = self._hitpoints[:, :2]
        center_xy = self._hitpoints[:, 6:8]
        dist = 1 - self._hitpoints[:, 9] / self._range

        for ind_pt in range(len(view_xy)):

            color = (
                int(dist[ind_pt] * 255),
                int(dist[ind_pt] * 255),
                int(dist[ind_pt] * 255),
            )
            arcade.draw_line(
                center_xy[ind_pt, 0],
                center_xy[ind_pt, 1],
                view_xy[ind_pt, 0],
                view_xy[ind_pt, 1],
                color,
            )

    @property
    def shape(self):
        return self._resolution, 1

    @property
    def _default_value(self):
        return np.zeros(self.shape)

    def _apply_normalization(self):
        self._values = self._values / self._range


class RGBSensor(RaySensor):
    def _compute_raw_sensor(self):
        self._values = self._hitpoints[:, 10:13]

    def draw(self):
        if self._disabled:
            return
        view_xy = self._hitpoints[:, :2]
        center_xy = self._hitpoints[:, 6:8]
        color = self._hitpoints[:, 10:13].astype(np.uint8)

        for ind_pt in range(len(view_xy)):

            if self._hitpoints[ind_pt, 8] == 0:
                continue
            color_pt = color[ind_pt]
            arcade.draw_line(
                center_xy[ind_pt, 0],
                center_xy[ind_pt, 1],
                view_xy[ind_pt, 0],
                view_xy[ind_pt, 1],
                color_pt,
            )

    @property
    def shape(self):
        return self._resolution, 3

    @property
    def _default_value(self):
        return np.zeros(self.shape)

    def _apply_normalization(self):
        self._values = self._values / 255.0


class SemanticSensor(RaySensor):
    def _compute_raw_sensor(self):
        self._values = self._hitpoints[:, 8:10]

    def draw(self):

        if self._disabled:
            return

        view_xy = self._hitpoints[:, :2]
        center_xy = self._hitpoints[:, 6:8]
        id_detection = self._hitpoints[:, 8].astype(np.int)

        for ind_pt in range(len(view_xy)):

            if id_detection[ind_pt] != 0:

                color = id_to_pixel(id_detection[ind_pt])

                arcade.draw_line(
                    center_xy[ind_pt, 0],
                    center_xy[ind_pt, 1],
                    view_xy[ind_pt, 0],
                    view_xy[ind_pt, 1],
                    color,
                )

    @property
    def shape(self):
        return self._resolution, 2

    @property
    def _default_value(self):
        return np.zeros(self.shape)

    def _apply_normalization(self):
        self._values = self._values / (1, self._range)



from .ray import DistanceSensor, RayCompute, RaySensor, RGBSensor, SemanticSensor
from .sensor import ExternalSensor, Sensor, SensorValue

__all__ = [
    "SensorValue",
    "Sensor",
    "ExternalSensor",
    "RGBSensor",
    "RayCompute",
    "RaySensor",
    "DistanceSensor",
    "SemanticSensor",
]



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Dict, List, Optional

import pymunk

from ...entity import EmbodiedEntity
from ...utils.definitions import CollisionTypes
from ..controller import GrasperController
from .interactor import ActiveInteractor

if TYPE_CHECKING:
    from ...entity import Graspable
    from ..part import PhysicalPart


class Grasper(ActiveInteractor, ABC):
    def __init__(
        self, anchor: PhysicalPart, max_grasped: Optional[int] = None, **kwargs
    ):
        super().__init__(anchor=anchor, **kwargs)

        self.grasp_controller = GrasperController("grasper")

        self._grasped_entities: List[EmbodiedEntity] = []

        self._grasp_joints: Dict[EmbodiedEntity, List[pymunk.PinJoint]] = {}

        self._can_grasp = False

        self._max_grasped = max_grasped

    @property
    def can_grasp(self):
        return self._can_grasp

    @property
    def grasped_entities(self):
        return self._grasped_entities

    def grasps(self, graspable: Graspable):

        assert self._can_grasp
        assert self._anchor

        entity = graspable.anchor

        if entity not in self._grasped_entities:

            if self._max_grasped and self._max_grasped <= len(self._grasped_entities):
                return

            self._grasped_entities.append(entity)
            self._add_joints(entity)
            entity.grasped_by.append(self)

            for sensor in self._anchor.agent.external_sensors:
                if sensor.invisible_grasped:
                    sensor.add_to_temporary_invisible(entity)

    @property
    def _collision_type(self):
        return CollisionTypes.GRASPER

    def _add_joints(self, entity: EmbodiedEntity):

        assert self._anchor

        j_1 = pymunk.PinJoint(self._anchor.pm_body, entity.pm_body, (0, 0), (0, 20))
        j_2 = pymunk.PinJoint(self._anchor.pm_body, entity.pm_body, (0, 0), (0, -20))

        j_3 = pymunk.PinJoint(self._anchor.pm_body, entity.pm_body, (0, 20), (0, 0))
        j_4 = pymunk.PinJoint(self._anchor.pm_body, entity.pm_body, (0, -20), (0, 0))

        grasp_joints = [j_1, j_2, j_3, j_4]
        self._grasp_joints[entity] = grasp_joints
        self._anchor.playground.space.add(*grasp_joints)

    def _release_grasping(self):

        for entity in list(self._grasped_entities):
            self.release(entity)

        self._can_grasp = False

        assert not self._grasped_entities
        assert not self._grasp_joints

    def release(self, entity):

        entity.grasped_by.remove(self)

        for sensor in self._anchor.agent.external_sensors:
            if sensor.invisible_grasped:
                sensor.remove_from_temporary_invisible(entity)

        joints = self._grasp_joints.pop(entity)
        self._anchor.playground.space.remove(*joints)

        self._grasped_entities.remove(entity)

    def reset(self):
        self._release_grasping()
        super().reset()


class GraspHold(Grasper):
    def __init__(self, anchor, **kwargs):

        super().__init__(anchor, **kwargs)

        self._is_grasping = False

    def apply_commands(self, **kwargs):
        command_value = self.grasp_controller.command_value

        if not command_value:
            self._release_grasping()
            self._is_grasping = False
        else:
            if not self._is_grasping:
                self._can_grasp = True
                self._is_grasping = True

            else:
                self._can_grasp = False


class GraspMagnet(Grasper):
    def apply_commands(self, **kwargs):
        command_value = self.grasp_controller.command_value

        if not command_value:
            self._release_grasping()
        else:
            self._can_grasp = True



from abc import ABC

from .sensor import ExternalSensor


class AreaSensor(ExternalSensor, ABC):
    """
    Base class for Area sensor.
    Area sensors use Pymunk shape detection to detect shapes within a certain radius.
    """



""" Contains the base class for entities.

Entity classes should be used to create body parts of
an agent, scene entities, spawners, timers, etc.

Entity is the generic building block of physical and interactive
objects in simple-playgrounds.

"""
from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, List, Optional, Union

if TYPE_CHECKING:
    from ..playground import Playground

Teams = Union[str, List[str]]


class Entity(ABC):
    """
    Base class that defines the entities that composing a Playground.
    Entities can be: SceneElement, Agent, Spawner, Timer, ...
    Entity can belong to one or multiple teams.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        teams: Optional[Teams] = None,
        temporary: bool = False,
        **_,
    ):

        # Unique identifiers
        self._uid = None
        self._name = name

        # Teams
        if isinstance(teams, str):
            teams = [teams]
        elif not teams:
            teams = []

        self._teams = teams

        self._temporary = temporary

        self._removed = False

        self._playground = None

    @property
    def uid(self):
        return self._uid

    @uid.setter
    def uid(self, uid: int):
        self._uid = uid

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def teams(self):
        return self._teams

    @property
    def removed(self):
        return self._removed

    @removed.setter
    def removed(self, rem: bool):
        self._removed = rem

    @property
    def temporary(self):
        return self._temporary

    @property
    def playground(self):
        return self._playground

    @playground.setter
    def playground(self, playground: Optional[Playground]):
        self._playground = playground

    @property
    def rng(self):

        if self._playground:
            return self._playground.rng

        return None

    # Interactions with playground

    def reset(self):
        """
        Upon reset of the Playgroung,
        revert the entity back to its original state.
        """

    def pre_step(self):
        """
        Preliminary calculations before the pymunk engine steps.
        """

    def post_step(self):
        """
        Updates the entity state after pymunk engine steps.
        """

    def update_team_filter(self):
        pass



"""
Module containing classes to generate random positions and trajectories

"""
from __future__ import annotations

import math
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Tuple, Union

import numpy as np

if TYPE_CHECKING:
    from ..playground import Playground

Coordinate = Tuple[Tuple[float, float], float]


class CoordinateSampler(ABC):
    def __init__(
        self,
        playground: Playground,
        center: Tuple[float, float],
        radius: Optional[float] = None,
        width: Optional[float] = None,
        height: Optional[float] = None,
        size: Optional[Tuple[float, float]] = None,
    ):

        self._radius = radius

        if (not width) and size:
            width, height = size

        self._width = width
        self._height = height

        self._center = center

        assert self._radius or self._width

        self._playground = playground

    def _get_relative_positions(self):

        if self._radius:
            pos = np.indices((self._radius, self._radius))
            dist = (pos[0, :] - self._radius / 2) ** 2 + (pos[1, :] - self._radius / 2)
            pos = np.where(dist < (self._radius / 2) ** 2)
            pos = pos - np.atleast_2d((self._radius / 2, self._radius / 2)).transpose()

        elif self._width:

            if self._height:
                pos = np.indices((self._width, self._height)).reshape(2, -1)
                pos = (
                    pos - np.atleast_2d((self._width / 2, self._height / 2)).transpose()
                )

            else:
                pos = np.indices((self._width, self._width)).reshape(2, -1)
                pos = (
                    pos - np.atleast_2d((self._width / 2, self._width / 2)).transpose()
                )

        return pos

    @property
    def _rng(self):
        return self._playground.rng

    @abstractmethod
    def _get_position_pdf(self, position_indices):
        ...

    def _get_random_angle(self):
        return self._rng.uniform(0, 2 * math.pi)

    def sample(self):
        """
        Sample probabiity for all possible coordinates,
        then sort them by order of posterior.
        """

        position_indices = self._get_relative_positions()
        position_pdf = self._get_position_pdf(position_indices)

        uniform_sampling = self._rng.uniform(size=position_pdf.size)
        posterior = uniform_sampling * position_pdf

        rr, cc = position_indices

        stacked = np.stack((rr, cc, posterior), axis=-1).reshape(-1, 3)
        sorted_coordinates = stacked[stacked[:, 2].argsort()]
        sorted_coordinates = sorted_coordinates[::-1]

        for rel_x, rel_y, _ in sorted_coordinates:

            x = self._center[0] + rel_x
            y = self._center[1] + rel_y
            angle = self._get_random_angle()

            yield (x, y), angle


class UniformCoordinateSampler(CoordinateSampler):
    def _get_position_pdf(self, position_indices):
        return np.ones(position_indices.shape[1])


class GaussianCoordinateSampler(CoordinateSampler):
    def __init__(self, playground, sigma, **kwargs):

        self._sigma = sigma
        super().__init__(playground, **kwargs)

    def _get_position_pdf(self, position_indices):
        rr, cc = position_indices
        return np.exp(-(rr**2 + cc**2) / self._sigma**2)


InitCoord = Union[Coordinate, CoordinateSampler]

geometric_shapes = {
    "line": 2,
    "circle": 60,
    "triangle": 3,
    "square": 4,
    "pentagon": 5,
    "hexagon": 6,
}



""" This module lists all the definitions.
"""

# pylint: disable=missing-class-docstring
from collections import namedtuple
from enum import IntEnum, auto

SIMULATION_STEPS = 10
SPACE_DAMPING = 0.9
LINEAR_FORCE = 100
ANGULAR_VELOCITY = 0.3
ARM_MAX_FORCE = 500
MAX_GRASP_FORCE = 600

WALL_DEPTH = 10

PYMUNK_STEPS = 10

VISIBLE_ALPHA = 255
INVISIBLE_ALPHA = 75
DEFAULT_INTERACTION_RANGE = 5


class PymunkCollisionCategories(IntEnum):

    # Must start at one:
    TRAVERSABLE = auto()
    TRANSPARENT = auto()
    SENSOR = auto()
    SENSOR_CONTACT = auto()
    INTERACTION = auto()
    BOUNDARY = auto()
    NO_TEAM = auto()


class KeyTypes(IntEnum):

    PRESS_HOLD = auto()
    PRESS_ONCE = auto()


FRICTION_ENTITY = 0.8
ELASTICITY_ENTITY = 0.5


class CollisionTypes(IntEnum):

    ELEMENT = auto()

    PART = auto()
    CONTACT = auto()
    ACTIVABLE = auto()
    GEM = auto()
    GRASPABLE = auto()
    ACTIVABLE_BY_GEM = auto()
    TELEPORT = auto()

    DEVICE = auto()
    DISABLER = auto()

    GRASPER = auto()
    ACTIVATOR = auto()

    # ACTIVE_INTERACTOR interacts only with PASSIVE_INTERACTOR
    # PASSIVE_INTERACTOR interacts with ACTIVE OR PASSIVE
    ACTIVE_INTERACTOR = auto()
    PASSIVE_INTERACTOR = auto()


def add_custom_collision(collision_types, name):
    """
    Function that allows to add new collisions to CollisionTypes.
    This is used when an user wants to create a new type of Entity,
    that requires particular collision handler and behavior.

    Examples:
        CustomCollisionTypes = add_custom_collision(CollisionTypes, JELLY)

    Args:
        collision_types: CollisionTypes enum.
            Either the default one or a custom one.
        name: name of the new collision type (UPPERCASE)

    Returns:
        New CollisionTypes Enum.

    """

    names = [m.name for m in collision_types] + [name]
    extended_enum = IntEnum("CollisionTypes", names)

    return extended_enum


class ElementTypes(IntEnum):

    # Basic Entities
    BASIC = auto()
    DOOR = auto()
    WALL = auto()
    COLOR_CHANGING = auto()

    # Contact entities
    VISIBLE_ENDGOAL = auto()
    VISIBLE_DEATHTRAP = auto()
    CANDY = auto()
    POISON = auto()
    SWITCH = ()

    # Edibles
    APPLE = auto()
    ROTTEN_APPLE = auto()

    # Gems
    # PASSIVE_INTERACTOR interacts with ACTIVE OR PASSIVE
    KEY = auto()

    # Interactive
    REWARD_ON_ACTIVATION = auto()
    DISPENSER = auto()
    CHEST = auto()
    VENDING_MACHINE = auto()
    LOCK = auto()

    # Zone
    GOAL_ZONE = auto()
    DEATH_ZONE = auto()
    TOXIC_ZONE = auto()
    HEALING_ZONE = auto()

    # Proximity
    FAIRY = auto()
    FIREBALL = auto()

    # Teleport
    BEAM = auto()
    BEAM_HOMING = auto()
    PORTAL = auto()

    # Modifier Zones
    SENSOR_DISABLER = auto()
    COMM_DISABLER = auto()
    ACTUATOR_DISABLER = auto()

    SPAWNER = auto()


class SensorTypes(IntEnum):

    SENSOR = auto()
    VISUAL = auto()
    SEMANTIC = auto()
    ROBOTIC = auto()

    RGB = auto()
    GREY = auto()
    LIDAR = auto()
    TOUCH = auto()
    Proximity = auto()

    SEMANTIC_CONE = auto()
    SEMANTIC_RAY = auto()
    TOPDOWN_LOCAL = auto()
    TOPDOWN_GLOBAL = auto()
    PERFECT_SEMANTIC = auto()


Detection = namedtuple("Detection", "entity, distance, angle")



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

if TYPE_CHECKING:
    from spg.playground.playground import Playground


class Timer(ABC):
    def __init__(self):

        self._running = False
        self._time = 0
        self._tic = False
        self._playground: Optional[Playground] = None

    @property
    def tic(self):
        return self._tic

    def start(self):
        self._running = True

    def stop(self):
        self._running = False

    def step(self):
        if self._running:
            self._time += 1

    def reset(self):
        self.stop()
        self._time = 0

    @property
    def playground(self):
        return self._playground

    @playground.setter
    def playground(self, pg):
        self._playground = pg

    @property
    def in_playground(self):
        return bool(self._playground)


class CountDownTimer(Timer):
    def __init__(self, duration: int):

        super().__init__()
        self._duration = duration

    def step(self):
        super().step()

        if self._tic:
            self._tic = False

        if self._time == self._duration:
            self.reset()
            self._tic = True


class PeriodicTimer(Timer):
    def __init__(self, durations: Union[List[int], int, Tuple[int, ...]]):

        if isinstance(durations, int):
            durations = [durations]

        assert isinstance(durations, (list, tuple))

        self._durations = durations
        self._current_duration = self._durations[0]
        self._index_duration = 0

        super().__init__()

    def reset(self):
        super().reset()
        self._index_duration = 0
        self._current_duration = self._durations[0]
        self._tic = False
        self.start()

    def step(self):

        super().step()

        if self.tic:
            self._tic = False

        if self._time == self._current_duration:
            self._index_duration = (self._index_duration + 1) % len(self._durations)
            self._current_duration = self._durations[self._index_duration]

            self._time = 0
            self._tic = True



import math

from .agent import Agent
from .communicator import Communicator
from .interactor import GraspHold
from .part import ForwardBase, Head
from .sensor import DistanceSensor, RGBSensor


class HeadAgent(Agent):
    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        base = ForwardBase(linear_ratio=10)
        self.add(base)

        self.head = Head(rotation_range=math.pi)
        base.add(self.head)

        # SENSORS
        self.distance = DistanceSensor(
            fov=360,
            resolution=36,
            max_range=100,
            invisible_elements=self._parts,
            normalize=True,
        )
        self.base.add(self.distance)

        self.rgb = RGBSensor(
            fov=180,
            resolution=64,
            max_range=400,
            invisible_elements=self._parts,
            invisible_when_grasped=True,
        )
        self.head.add(self.rgb)

        # COMMS
        self.comm = Communicator()
        self.base.add(self.comm)

        # Grapser
        grasp = GraspHold(base)
        self.base.add(grasp)



"""
Module that defines Base Class SceneElement
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

from ..entity import (
    Entity,
    Graspable,
    InteractiveAnchored,
    InteractiveZone,
    PhysicalEntity,
)
from ..utils.definitions import CollisionTypes


class SceneElement(ABC):
    pass


class PhysicalElement(PhysicalEntity, SceneElement):
    def __init__(self, **entity_params):
        super().__init__(**entity_params)

        self._interactives = []

        self._grasped_by = []
        self._graspable = False

        self._produced_by: Optional[Entity] = None

    @property
    def graspable(self):
        return bool(self._graspable)

    @graspable.setter
    def graspable(self, graspable: bool):
        if graspable and not self._graspable:
            self._graspable = Graspable(anchor=self)
            self.add(self._graspable)
        elif not graspable and self._graspable:
            assert self._playground
            self._playground.remove(self._graspable)
            self._graspable = False

    @property
    def grasped_by(self):
        return self._grasped_by

    @property
    def produced_by(self):
        return self._produced_by

    @property
    def interactives(self):
        return self._interactives

    def add(self, interactive: InteractiveAnchored):

        self._interactives.append(interactive)
        interactive.pm_body = self._pm_body
        interactive.teams = self._teams
        interactive.anchor = self

        if self._playground:
            self._playground.add(interactive)

    @property
    def _collision_type(self):
        return CollisionTypes.ELEMENT

    def update_team_filter(self):
        super().update_team_filter()

        for interactive in self._interactives:
            interactive.update_team_filter()

    # def update_sprite(self, view, sprite, force=False):
    #     super().update_sprite(view, sprite, force)

    # for interactive in self._interactives:
    #     interactive.update_sprite(view, sprite, force)

    # TODO: pre step, reset etc


class ZoneElement(InteractiveZone, SceneElement):
    pass


class RewardElement(ABC):
    @property
    @abstractmethod
    def _base_reward(self) -> float:
        ...

    @property
    def reward(self) -> float:
        return self._base_reward


class Teleporter(ZoneElement):
    pass



import math
from abc import ABC

import numpy as np

from .sensor import Sensor

##################
# Internal Sensors
##################


class InternalSensor(Sensor, ABC):

    """
    Base Class for Internal Sensors.
    """


class Position(InternalSensor):

    """
    Position Sensor returns a numpy array containing the position x, y and the
    orientation of the anchor.
    """

    def _compute_raw_sensor(self):
        self._values = np.asarray((*self._anchor.position, self._anchor.angle))

    def _apply_normalization(self):

        assert self._playground

        if self._playground.size:
            self._values /= (
                self._playground.size[0],
                self._playground.size[1],
                2 * math.pi,
            )

    @property
    def shape(self):
        return (3, 1)


class Velocity(InternalSensor):
    def _compute_raw_sensor(self):
        self._values = np.asarray(
            (*self._anchor.velocity, self._anchor.angular_velocity)
        )

    def _apply_normalization(self):
        pass

    @property
    def shape(self):
        return (3, 1)


class RelativeVelocity(InternalSensor):
    def _compute_raw_sensor(self):

        vel = self._anchor.velocity.rotated(-self._anchor.angle)

        self._values = np.asarray((*vel, self._anchor.angular_velocity))

    def _apply_normalization(self):
        pass

    @property
    def shape(self):
        return (3, 1)


class Time(InternalSensor):
    def _compute_raw_sensor(self):
        self._values = np.array(self.playground.timestep)

    def _apply_normalization(self):
        pass

    @property
    def shape(self):
        return (1, 1)



# pylint: disable=too-many-public-methods

from __future__ import annotations

import math
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Union

import pymunk
from arcade import Sprite, Texture
from PIL import Image

from ..utils.definitions import ELASTICITY_ENTITY, FRICTION_ENTITY
from ..utils.position import Coordinate, CoordinateSampler, InitCoord
from .entity import Entity

Teams = Union[str, List[str]]


class EmbodiedEntity(Entity, ABC):

    """
    Embodied Entities are elementary entities physically present in the playground.
    They can interact with other Embodied Entities through
    pymunk collisions.

    """

    def __init__(
        self,
        name: Optional[str] = None,
        filename: Optional[str] = None,
        texture: Optional[Texture] = None,
        radius: Optional[float] = None,
        width: Optional[float] = None,
        height: Optional[float] = None,
        temporary: bool = False,
        shape_approximation: Optional[str] = None,
        sprite_front_is_up: bool = False,
        teams: Optional[Teams] = None,
        color: Optional[Tuple[int, int, int]] = None,
    ):

        super().__init__(name, teams, temporary)

        # Get the base sprite
        self._sprite_front_is_up = sprite_front_is_up
        self._base_sprite = Sprite(
            texture=texture,
            filename=filename,
            hit_box_algorithm="Detailed",
            hit_box_detail=1,
            flipped_diagonally=sprite_front_is_up,
            flipped_horizontally=sprite_front_is_up,
        )  # type: ignore

        self._color = color

        # Get the scale and dimensions
        self._scale, self._radius, self._width, self._height = self._get_dimensions(
            radius, width, height
        )

        # Get pymunk elements
        self._pm_body = self._get_pm_body()
        self._pm_shapes = self._get_pm_shapes_from_sprite(shape_approximation)

        # Interactions and collisions
        self._set_pm_collision_type()

        # Flags
        self._moved = False

        # At the begining, not in playground
        self._allow_overlapping = False
        self._initial_coordinates: Optional[InitCoord] = None

    #############
    # Properties
    #############
    @property
    def pm_elements(self):
        return self._pm_shapes + [self._pm_body]

    @property
    def pm_shapes(self):
        return self._pm_shapes

    @property
    def pm_body(self):
        return self._pm_body

    @property
    def texture(self):
        return self._base_sprite.texture

    @property
    def radius(self):
        return self._radius

    @property
    def scale(self):
        return self._scale

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def coordinates(self):
        return self.position, self.angle

    @property
    def position(self):
        """Position (x, y)"""
        return self._pm_body.position

    @property
    def angle(self):
        """Absolute orientation"""
        return self._pm_body.angle % (2 * math.pi)

    @property
    def velocity(self):
        return self._pm_body.velocity

    @property
    def angular_velocity(self):
        return self._pm_body.angular_velocity

    @property
    def color_uid(self):
        return self._uid & 255, (self._uid >> 8) & 255, (self._uid >> 16) & 255, 255

    @property
    def moved(self):

        if self._moved:
            return True

        if self._pm_body.body_type == pymunk.Body.DYNAMIC:

            vel = self._pm_body.velocity.length
            if vel > 0.001:
                return True

            ang_vel = self._pm_body.angular_velocity
            if abs(ang_vel) > 0.001:
                return True

        return False

    @property
    def allow_overlapping(self):
        return self._allow_overlapping

    @allow_overlapping.setter
    def allow_overlapping(self, allow: bool):
        self._allow_overlapping = allow

    @property
    def initial_coordinates(self):
        return self._initial_coordinates

    @initial_coordinates.setter
    def initial_coordinates(self, init_coord: InitCoord):
        self._initial_coordinates = init_coord

    ##############
    # Init pm Elements
    ###############
    def _get_dimensions(self, radius, width, height):

        orig_radius = max(
            pymunk.Vec2d(*vert).length for vert in self._base_sprite.get_hit_box()
        )

        horiz = [pymunk.Vec2d(*vert).x for vert in self._base_sprite.get_hit_box()]
        vert = [pymunk.Vec2d(*vert).y for vert in self._base_sprite.get_hit_box()]

        orig_width = max(horiz) - min(horiz)
        orig_height = max(vert) - min(vert)

        # If radius imposed:
        if radius:
            scale = radius / orig_radius

        elif height:
            scale = height / orig_height

        elif width:
            scale = width / orig_width
        else:
            scale = 1

        width = scale * orig_width
        height = scale * orig_height
        radius = scale * orig_radius

        return scale, radius, width, height

    def _get_pm_shapes_from_sprite(self, shape_approximation):

        vertices = self._base_sprite.get_hit_box()

        vertices = [(x * self._scale, y * self._scale) for x, y in vertices]

        if shape_approximation == "circle":
            pm_shapes = [pymunk.Circle(self._pm_body, self._radius)]

        elif shape_approximation == "box":
            top = max(vert[0] for vert in vertices)
            bottom = min(vert[0] for vert in vertices)
            left = min(vert[1] for vert in vertices)
            right = max(vert[1] for vert in vertices)

            box_vertices = ((top, left), (top, right), (bottom, right), (bottom, left))

            pm_shapes = [pymunk.Poly(self._pm_body, box_vertices)]

        elif shape_approximation == "hull":
            pm_shapes = [pymunk.Poly(self._pm_body, vertices)]

        elif shape_approximation == "decomposition":

            if not pymunk.autogeometry.is_closed(vertices):
                vertices += [vertices[0]]

            if pymunk.area_for_poly(vertices) < 0:
                vertices = list(reversed(vertices))

            list_vertices = pymunk.autogeometry.convex_decomposition(
                vertices, tolerance=0.5
            )

            pm_shapes = []
            for vertices in list_vertices:
                pm_shape = pymunk.Poly(body=self._pm_body, vertices=vertices)
                pm_shapes.append(pm_shape)

        else:
            pm_shapes = [pymunk.Poly(body=self._pm_body, vertices=vertices)]

        for pm_shape in pm_shapes:
            pm_shape.friction = FRICTION_ENTITY
            pm_shape.elasticity = ELASTICITY_ENTITY

        return pm_shapes

    ##############
    # Sprites
    ##############

    def get_sprite(self, zoom: float = 1, color_uid=False) -> Sprite:

        texture = self._base_sprite.texture
        if color_uid:
            texture = self.color_with_id(texture)

        assert isinstance(texture, Texture)

        sprite = Sprite(
            texture=texture,
            scale=zoom * self._scale,
            hit_box_algorithm="Detailed",
            hit_box_detail=1,
        )

        if self._color and not color_uid:
            sprite.color = self._color

        return sprite

    def color_with_id(self, texture) -> Texture:

        img_uid = Image.new("RGBA", texture.size)
        pixels = img_uid.load()
        pixels_texture = texture.image.load()

        for i in range(img_uid.size[0]):
            for j in range(img_uid.size[1]):

                if pixels_texture[i, j][3] == 0:
                    pixels[i, j] = (0, 0, 0, 0)

                elif pixels_texture[i, j][:3] != (0, 0, 0):  # type: ignore
                    pixels[i, j] = self.color_uid  # type: ignore

                else:
                    pixels[i, j] = (0, 0, 0, 0)

        texture = Texture(
            name=str(self._uid),
            image=img_uid,
            hit_box_algorithm="Detailed",
            hit_box_detail=1,
        )

        return texture

    @property
    def needs_sprite_update(self):
        return self._moved

    def update_sprite(self, view, sprite):

        pos_x = (
            self._pm_body.position.x - view.center[0]
        ) * view.zoom + view.width // 2
        pos_y = (
            self._pm_body.position.y - view.center[1]
        ) * view.zoom + view.height // 2

        sprite.set_position(pos_x, pos_y)
        sprite.angle = int(self._pm_body.angle * 180 / math.pi)

    ###################
    # Pymunk objects
    ###################

    @abstractmethod
    def _get_pm_body(self) -> pymunk.Body:
        """
        Set pymunk body. Shapes are attached to a body.
        """

    def _set_pm_collision_type(self):
        for pm_shape in self._pm_shapes:
            pm_shape.collision_type = self._collision_type

    @property
    @abstractmethod
    def _collision_type(self):
        """
        Set the collision type for the pm shapes.
        """

    def update_team_filter(self):

        for pm_shape in self._pm_shapes:
            categ = pm_shape.filter.categories

            for team_name in self._teams:
                categ = categ | 2 ** self._playground.teams[team_name]

            mask = pm_shape.filter.mask
            for team_name, team_id in self._playground.teams.items():

                if team_name not in self._teams:
                    mask = mask | 2**team_id ^ 2**team_id

            pm_shape.filter = pymunk.ShapeFilter(categories=categ, mask=mask)

    ###################
    # Position and Move
    ###################
    def move_to(
        self,
        coordinates: Union[Coordinate, CoordinateSampler],
        allow_overlapping: bool = True,
        check_within: bool = False,
    ):

        assert self._playground

        if isinstance(coordinates, CoordinateSampler):
            coordinates = self._sample_valid_coordinate()

        if (not allow_overlapping) and self._playground.overlaps(self, coordinates):
            raise ValueError("Entity overlaps but overlap is not allowed")

        if check_within and not self._playground.within_playground(
            coordinates=coordinates
        ):
            raise ValueError("Entity is not placed within Playground boundaries")

        position, angle = coordinates

        # Calculate Velocities
        absolute_velocity, angular_velocity = (0, 0), 0

        # Apply new position and velocities
        self._pm_body.position, self._pm_body.angle = position, angle
        self._pm_body.velocity = absolute_velocity
        self._pm_body.angular_velocity = angular_velocity

        if self._pm_body.space:
            self._pm_body.space.reindex_shapes_for_body(self._pm_body)

        self._moved = True

    def _sample_valid_coordinate(self) -> Coordinate:

        assert self._playground

        sampler = self._initial_coordinates
        assert isinstance(sampler, CoordinateSampler)

        for coordinate in sampler.sample():
            if not self._playground.overlaps(self, coordinate):
                return coordinate

        raise ValueError("Entity could not be placed without overlapping")

    ##############
    # PLayground Interactions
    ################

    def pre_step(self):
        self._moved = False



from .controller import (
    BoolController,
    CenteredContinuousController,
    Command,
    ContinuousController,
    Controller,
    DiscreteController,
    GrasperController,
    RangeController,
)

__all__ = [
    "BoolController",
    "CenteredContinuousController",
    "Command",
    "ContinuousController",
    "Controller",
    "DiscreteController",
    "RangeController",
    "GrasperController",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from ..agent.device import Device
from ..agent.interactor import Grasper
from ..agent.part import PhysicalPart
from ..element import Disabler

if TYPE_CHECKING:
    from .playground import Playground


def get_colliding_entities(playground: Playground, arbiter):

    entity_1 = playground.get_entity_from_shape(arbiter.shapes[0])
    entity_2 = playground.get_entity_from_shape(arbiter.shapes[1])

    agent_1 = None
    agent_2 = None

    if isinstance(entity_1, PhysicalPart):
        agent_1 = entity_1.agent

    if isinstance(entity_2, PhysicalPart):
        agent_2 = entity_2.agent

    return (entity_1, agent_1), (entity_2, agent_2)


# Collision Handlers


def grasper_grasps_graspable(arbiter, _, data):

    playground: Playground = data["playground"]
    (grasper, _), (entity, _) = get_colliding_entities(playground, arbiter)

    assert isinstance(grasper, Grasper)

    if grasper.can_grasp:
        grasper.grasps(entity)

    return True


def disabler_disables_device(arbiter, _, data):

    playground: Playground = data["playground"]
    (disabler, _), (device, _) = get_colliding_entities(playground, arbiter)

    assert isinstance(device, Device)
    assert isinstance(disabler, Disabler)

    disabler.disable(device)

    return True


# def agent_touches_element(arbiter, space, data):

#     playground: Playground = data['playground']
#     (part, agent), (touched_element, _) = get_colliding_entities(playground, arbiter)

#     if not touched_element:
#         return True

#     assert isinstance(touched_element, InteractiveElement)
#     assert isinstance(agent, Agent)

#     agent.reward += touched_element.reward

#     elems_remove, elems_add = touched_element.activate(agent)
#     playground.remove_add_within(elems_remove, elems_add)

#     if touched_element.terminate_upon_activation:
#         playground.done = True

#     return True


# def agent_activates_element(arbiter, space, data):

#     playground: Playground = data['playground']
#     (part, agent), (activable_element, _) =
# get_colliding_entities(playground, arbiter)

#     if not activable_element:
#         return True

#     assert isinstance(activable_element, InteractiveElement)
# def grasper_grasps_graspable(arbiter, space, data)
#     assert isinstance(agent, Agent)

#     # Note: later, should handle the case where two agents activate simultaneously.
#     for actuator in agent.actuators:
#         if actuator.part is part and isinstance(actuator, Activate):

#             if actuator.is_activating and not activable_element.activated:

#                 agent.reward += activable_element.reward

#                 elems_remove, elems_add = activable_element.activate(agent)
#                 playground.remove_add_within(elems_remove, elems_add)

#                 if activable_element.terminate_upon_activation:
#                     playground.done = True

#                 actuator.is_activating = False

#     return True


# def gem_activates_element(arbiter, space, data):

#     playground: Playground = data['playground']
#     (gem, _), (activable_element, _) = get_colliding_entities(playground, arbiter)

#     assert isinstance(activable_element, InteractiveElement)

#     agent = playground.get_closest_agent(gem)

#     if not gem:
#         return True

#     assert isinstance(gem, GemElement)

#     if not activable_element.activated:

#         elems_remove, elems_add = activable_element.activate(gem)
#         playground.remove_add_within(elems_remove, elems_add)

#         agent.reward += activable_element.reward

#         if activable_element.terminate_upon_activation:
#             playground.done = True

#     return True


# def agent_teleports(arbiter, space, data):

#     playground: Playground = data['playground']
#     (_, agent), (teleport, _) = get_colliding_entities(playground, arbiter)

#     assert isinstance(teleport, TeleportElement)

#     if agent.has_teleported:
#         return True

#     teleport.energize(agent)

#     return True



from typing import Tuple

from pymunk import Vec2d

from ..element.wall import ColorWall
from ..playground import Playground
from ..utils.position import UniformCoordinateSampler


class ConnectedRooms(Playground):
    def __init__(
        self,
        size_room: Tuple[int, int],
        room_layout: Tuple[int, int],
        doorstep_length: float,
        centered_doorstep: bool = True,
        seed=None,
        background=None,
        wall_cls=ColorWall,
        wall_color=None,
        **kwargs,
    ):

        size = (size_room[0] * room_layout[0], size_room[1] * room_layout[1])

        self._size_room = size_room
        self._room_layout = room_layout

        self._doorstep_length = doorstep_length
        self._centered_doorstep = centered_doorstep
        self._color = wall_color

        super().__init__(size, seed, background, **kwargs)

        self._room_coordinates = self._get_room_coordinates()
        self._room_centers, self._room_corners = self._get_room_positions()

        self._wall_cls = wall_cls
        self._doorsteps = self._add_walls()

        self._room_coordinate_sampler = self._get_coord_sampler()

    def _get_room_coordinates(self):

        room_coordinates = []

        for x in range(self._room_layout[0]):
            for y in range(self._room_layout[1]):
                room_coordinates.append((x, y))

        return room_coordinates

    def _get_room_positions(self):

        room_centers = []
        room_corners = []

        corner_x, corner_y = -self.size[0] / 2, -self.size[1] / 2

        for (ind_x, ind_y) in self._room_coordinates:
            l_x = corner_x + ind_x * self._size_room[0]
            r_x = corner_x + (ind_x + 1) * self._size_room[0]
            b_y = corner_y + ind_y * self._size_room[1]
            t_y = corner_y + (ind_y + 1) * self._size_room[1]

            c_x = corner_x + (ind_x + 0.5) * self._size_room[0]
            c_y = corner_y + (ind_y + 0.5) * self._size_room[1]

            room_centers.append((c_x, c_y))
            room_corners.append(((l_x, b_y), (l_x, t_y), (r_x, t_y), (r_x, b_y)))

        return room_centers, room_corners

    def _add_walls(self):

        doorsteps = {}

        for (ind_x, ind_y), corners in zip(self._room_coordinates, self._room_corners):

            corn_bl, corn_tl, corn_tr, corn_br = corners

            if ind_x == 0:
                self._add_wall(corn_bl, corn_tl)

            if ind_x == self._room_layout[0] - 1:
                self._add_wall(corn_br, corn_tr)
            else:
                doorstep_pos = self._add_wall(corn_br, corn_tr, doorstep=True)
                if doorstep_pos:
                    doorsteps[(ind_x, ind_y), (ind_x + 1, ind_y)] = doorstep_pos

            if ind_y == 0:
                self._add_wall(corn_bl, corn_br)

            if ind_y == self._room_layout[1] - 1:
                self._add_wall(corn_tl, corn_tr)
            else:
                doorstep_pos = self._add_wall(corn_tl, corn_tr, doorstep=True)
                if doorstep_pos:
                    doorsteps[(ind_x, ind_y), (ind_x, ind_y + 1)] = doorstep_pos

        return doorsteps

    def _add_wall(self, pos_1, pos_2, doorstep=False):

        if not self._color:
            color = list(self._rng.integers(0, 255, 3))
        else:
            color = self._color

        if not doorstep:
            wall = self._wall_cls(pos_1, pos_2, width=10, color=color)
            self.add(wall, wall.wall_coordinates)
            return False

        pt_1 = Vec2d(*pos_1)
        pt_2 = Vec2d(*pos_2)
        angle = (pt_2 - pt_1).angle
        unit_vec = (pt_2 - pt_1).normalized()
        length = (pt_2 - pt_1).length

        if self._centered_doorstep:
            pt_doorstep = (pt_2 + pt_1) / 2
            pos = length / 2

        else:
            pos = self._rng.uniform(
                self._doorstep_length / 2, length - self._doorstep_length / 2
            )

        pt_doorstep = pt_1 + unit_vec * pos

        pt_ds_1 = pt_1 + unit_vec * (pos - self._doorstep_length / 2)
        pt_ds_2 = pt_1 + unit_vec * (pos + self._doorstep_length / 2)

        wall = self._wall_cls(pos_1, pt_ds_1, width=10, color=color)
        self.add(wall, wall.wall_coordinates)

        wall = self._wall_cls(pt_ds_2, pos_2, width=10, color=color)
        self.add(wall, wall.wall_coordinates)

        return (pt_doorstep.x, pt_doorstep.y), angle

    def _get_coord_sampler(self):

        samplers = []

        for center in self._room_centers:

            sampler = UniformCoordinateSampler(
                self, center, width=self._size_room[0], height=self._size_room[1]
            )
            samplers.append(sampler)

        return samplers


class Room(ConnectedRooms):
    def __init__(
        self,
        size: Tuple[int, int],
        seed=None,
        background=None,
        wall_cls=ColorWall,
        wall_color=None,
        **kwargs,
    ):

        super().__init__(
            size_room=size,
            room_layout=(1, 1),
            doorstep_length=10,
            centered_doorstep=True,
            seed=seed,
            background=background,
            wall_cls=wall_cls,
            wall_color=wall_color,
            **kwargs,
        )



from pathlib import Path

from arcade.resources import add_resource_handle

#: The absolute path to this directory
SPG_RESOURCE_PATH = Path(__file__).parent.resolve()

add_resource_handle("spg", SPG_RESOURCE_PATH)



""" Module that defines the base class Part and Actuator.

Part inherits from Entity, and is used to create different body parts
of an agent. Parts are visible and movable by default.

Examples on how to add Parts to an agent can be found
in spg/agents/agents.py
"""
from __future__ import annotations

import math
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import pymunk

from ...entity import PhysicalEntity
from ...utils.definitions import CollisionTypes
from ..controller import Command, Controller
from ..device import Device
from ..interactor.grasper import Grasper
from ..interactor.interactor import ActiveInteractor
from ..sensor import RaySensor

if TYPE_CHECKING:
    from ...utils.position import Coordinate
    from ..agent import Agent

CommandDict = Dict[Command, Union[float, int]]


class PhysicalPart(PhysicalEntity, ABC):
    def __init__(self, **kwargs):

        super().__init__(
            teams=None,
            **kwargs,
        )

        # Add physical motors if needed
        self._anchored: List[AnchoredPart] = []
        self._devices: List[Device] = []

        self._agent: Optional[Agent] = None

        self.grasper_controller: Optional[Controller] = None
        self.grasper: Optional[Grasper] = None

    @property
    def agent(self):
        return self._agent

    @agent.setter
    def agent(self, agent: Agent):
        self._agent = agent

    @property
    def anchored(self):
        return self._anchored

    @property
    def devices(self):
        return self._devices

    @property
    def teams(self):
        return self._teams

    @teams.setter
    def teams(self, teams):
        self._teams = teams

    def add(
        self,
        elem: Union[AnchoredPart, Device],
        anchor_coordinates: Optional[Coordinate] = None,
    ):

        elem.anchor = self
        elem.teams = self._teams

        if isinstance(elem, AnchoredPart):
            if anchor_coordinates:
                elem.anchor_coordinates = anchor_coordinates
            self._anchored.append(elem)

        elif isinstance(elem, Device):
            self._devices.append(elem)

            if isinstance(elem, RaySensor) and self._playground:
                self.playground.sensor_shader.add(elem)

            elif isinstance(elem, Grasper):
                if self.grasper:
                    raise ValueError("Grasper already in. Only one grasper per part.")

                self.grasper = elem
                self.grasper_controller = elem.grasp_controller
                self.add(self.grasper_controller)

        else:
            raise ValueError("Not implemented")

        if isinstance(elem, AnchoredPart):
            assert self._agent
            self._agent.add(elem)

    def move_to(  # pylint: disable=arguments-differ
        self,
        coordinates: Coordinate,
        move_anchors=False,
        **kwargs,
    ):
        super().move_to(coordinates=coordinates, **kwargs)

        if move_anchors:
            for part in self._anchored:
                part_coord = part.get_init_coordinates()
                part.move_to(
                    coordinates=part_coord, move_anchors=move_anchors, **kwargs
                )

    @property
    def _collision_type(self):
        return CollisionTypes.PART

    def update_team_filter(self):
        super().update_team_filter()

        for device in self._devices:
            device.update_team_filter()

    def apply_commands(self, **kwargs):

        self._apply_commands(**kwargs)

        for device in self.devices:
            if isinstance(device, ActiveInteractor):
                device.apply_commands(**kwargs)

    @abstractmethod
    def _apply_commands(self, **kwargs):
        ...

    def pre_step(self):
        super().pre_step()
        for device in self._devices:
            device.pre_step()

    def post_step(self):
        super().post_step()
        for device in self._devices:
            device.post_step()

    def reset(self):
        super().reset()

        self._pm_body.velocity = (0, 0)
        self._pm_body.angular_velocity = 0

        for part in self._anchored:
            part.reset()

        for device in self._devices:
            device.reset()


class AnchoredPart(PhysicalPart, ABC):
    def __init__(
        self,
        rotation_range: float,
        **kwargs,
    ):

        super().__init__(
            **kwargs,
        )

        self._rotation_range = rotation_range

        self._anchor = None
        self._motor = None
        self._joint = None
        self._limit = None

        # Point on the anchor
        self._anchor_coordinates = ((0, 0), 0)

    @property
    def anchor(self):
        return self._anchor

    @anchor.setter
    def anchor(self, anchor):
        self._anchor = anchor

    # @property
    # def agent(self):
    # assert self._anchor
    # return self._anchor.agent

    @property
    def anchor_coordinates(self):
        return self._anchor_coordinates

    @anchor_coordinates.setter
    def anchor_coordinates(self, coord: Coordinate):
        self._anchor_coordinates = coord

    @property
    def relative_position(self):
        assert self._anchor
        return (self.position - self._anchor.position).rotated(-self._anchor.angle)

    @property
    def relative_angle(self):
        assert self._anchor
        rel_angle = (self.angle - self._anchor.angle) % (2 * math.pi)

        if rel_angle > math.pi:
            return rel_angle - 2 * math.pi
        return rel_angle

    def get_init_coordinates(self):
        """
        Calculates the position of a Part relative to its Anchor.
        Sets the position of the Part.
        """

        assert self._anchor

        pos_anchor = pymunk.Vec2d(*self._anchor_coordinates[0])
        pos_pivot = pymunk.Vec2d(*self._pivot_position)

        angle_offset_anchor = self._anchor_coordinates[1]

        position = (
            self._anchor.position
            + pos_anchor.rotated(self._anchor.angle)
            - pos_pivot.rotated(self._anchor.angle + angle_offset_anchor)
        )

        angle = self._anchor.pm_body.angle + angle_offset_anchor

        return position, angle

    def attach_to_anchor(self):

        assert self._anchor

        angle_offset_anchor = self._anchor_coordinates[1]

        # Create joint to attach to anchor
        self._joint = pymunk.PivotJoint(
            self._anchor.pm_body,
            self.pm_body,
            self._anchor_coordinates[0],
            self._pivot_position,
        )
        self._joint.collide_bodies = False
        self._limit = pymunk.RotaryLimitJoint(
            self._anchor.pm_body,
            self._pm_body,
            angle_offset_anchor - self._rotation_range / 2,
            angle_offset_anchor + self._rotation_range / 2,
        )

        self._motor = pymunk.SimpleMotor(self._anchor.pm_body, self.pm_body, 0)

    @property
    def pm_joints(self):
        assert self._joint and self._limit and self._motor
        return self._joint, self._limit, self._motor

    @property
    @abstractmethod
    def _pivot_position(self) -> pymunk.Vec2d:
        ...



from __future__ import annotations

import math
from abc import ABC, abstractmethod

import numpy as np
import pymunk
from arcade.resources import resolve_resource_path
from arcade.texture import Texture
from PIL import Image
from skimage.io import imread
from skimage.transform import resize

from spg.element.element import PhysicalElement


class Wall(PhysicalElement):
    def __init__(self, pos_start, pos_end, width, color, **kwargs):

        length = (pymunk.Vec2d(*pos_start) - pos_end).length

        position = (pymunk.Vec2d(*pos_start) + pos_end) / 2
        angle = (pymunk.Vec2d(*pos_end) - pos_start).angle + math.pi / 2

        self.wall_coordinates = position, angle

        img = self._get_img_wall(width, length)

        texture = Texture(
            name=f"{type(self)}_{width}_{length}_{color}",
            image=img,
            hit_box_algorithm="None",
            hit_box_detail=1,
        )

        super().__init__(texture=texture, color=color, **kwargs)

    @property
    def length(self):
        return self._height

    @abstractmethod
    def _get_img_wall(self, width, length):
        ...


class ColorWall(Wall):
    def _get_img_wall(self, width, length):
        return Image.new("RGBA", (int(width), int(length)), (255, 255, 255, 255))


class TiledColorWall(Wall, ABC):
    _fname: str

    def _get_img_wall(self, width, length):

        file_name = resolve_resource_path(self._fname)

        tile_array = imread(file_name)

        length_resize = tile_array.shape[1] * width / tile_array.shape[0]
        tile_array = resize(tile_array, (width, length_resize))

        n_repeats = int(length / tile_array.shape[0]) + 1
        new_img = np.tile(tile_array, reps=(1, n_repeats, 1))
        new_img = (new_img[:, : int(length), :] * 255).astype(np.uint8)
        new_img = new_img.swapaxes(1, 0)

        return Image.fromarray(new_img)


class TiledAlternateColorWall(TiledColorWall):
    _fname = ":spg:spg/tile_alternate.png"


class TiledGradientColorWall(TiledColorWall):
    _fname = ":spg:spg/tile_grad.png"


class TiledLongColorWall(TiledColorWall):
    _fname = ":spg:spg/tile_long.png"



from typing import Optional, Tuple

from .element import PhysicalElement


class Ball(PhysicalElement):
    def __init__(self, color: Optional[Tuple[int, int, int]] = None):

        super().__init__(
            mass=10,
            filename=":spg:rollingball/ball/ball_blue_large.png",
            radius=10,
            color=color,
        )

    def _set_pm_collision_type(self):
        pass



from __future__ import annotations

from abc import ABC
from typing import Optional

import pymunk

from ..utils.definitions import PymunkCollisionCategories
from .embodied import EmbodiedEntity


class PhysicalEntity(EmbodiedEntity, ABC):
    """
    PhysicalEntity creates a physical object that can collide with
    other Physical Entities.
    It deals with physical properties such as the mass, visual texture,
     whether it is transparent or traversable.
    PhysicalEntity is visible and non-traversable by default.
    An agent is composed of multiple PhysicalEntity that are attached to each other.
    """

    def __init__(
        self,
        mass: Optional[float] = None,
        traversable: bool = False,
        transparent: bool = False,
        **kwargs,
    ):

        self._mass = mass
        self._transparent = transparent
        self._traversable = traversable

        super().__init__(**kwargs)

        self._set_shape_collision_filter()

    @property
    def transparent(self):
        return self._transparent

    @property
    def traversable(self):
        return self._traversable

    @property
    def movable(self):
        return bool(self._mass)

    @property
    def needs_sprite_update(self):
        return self._moved or self.movable

    ########################
    # BODY AND SHAPE
    ########################

    def _get_pm_body(self):

        if not self._mass:
            return pymunk.Body(body_type=pymunk.Body.STATIC)

        vertices = self._base_sprite.get_hit_box()
        moment = pymunk.moment_for_poly(self._mass, vertices)

        return pymunk.Body(self._mass, moment, body_type=pymunk.Body.DYNAMIC)

    def _set_shape_collision_filter(self):

        # By default, a physical entity collides with all
        if self._transparent and self._traversable:
            raise ValueError(
                "Physical Entity can not be transparent and traversable.\
                Use Interactive Entity."
            )

        categories = 2**PymunkCollisionCategories.NO_TEAM.value
        mask = (
            pymunk.ShapeFilter.ALL_MASKS()
            ^ 2**PymunkCollisionCategories.TRAVERSABLE.value
        )

        # If traversable, collides with nothing
        if self._traversable:
            categories = 2**PymunkCollisionCategories.TRAVERSABLE.value
            mask = 0

        # If transparent, collides with everything except tra.
        # if self._transparent:
        #     categories = 2 ** PymunkCollisionCategories.TRANSPARENT.value

        for pm_shape in self._pm_shapes:
            pm_shape.filter = pymunk.ShapeFilter(categories=categories, mask=mask)



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
from arcade import SpriteList
from arcade.sprite import Sprite

from ..entity import InteractiveAnchored, InteractiveZone, PhysicalEntity

if TYPE_CHECKING:
    from ..entity import EmbodiedEntity
    from ..playground import Playground


class TopDownView:
    def __init__(
        self,
        playground: Playground,
        size: Optional[Tuple[int, int]] = None,
        center: Tuple[float, float] = (0, 0),
        zoom: float = 1,
        display_uid: bool = False,
        draw_transparent: bool = True,
        draw_interactive: bool = True,
        draw_zone: bool = True,
    ) -> None:

        self._playground = playground

        self._ctx = playground.window.ctx

        self._center = center

        if not size:
            size = playground.size

        if not size:
            raise ValueError("Size should be set")

        self._width, self._height = self._size = size

        self._zoom = zoom
        self._draw_transparent = draw_transparent
        self._draw_interactive = draw_interactive
        self._draw_zone = draw_zone
        self._display_uid = display_uid

        self._transparent_sprites = SpriteList()
        self._visible_sprites = SpriteList()
        self._interactive_sprites = SpriteList()
        self._zone_sprites = SpriteList()
        self._traversable_sprites = SpriteList()

        self._background = playground.background

        self._fbo = self._ctx.framebuffer(
            color_attachments=[
                self._ctx.texture(
                    (size),
                    components=4,
                    wrap_x=self._ctx.CLAMP_TO_BORDER,  # type: ignore
                    wrap_y=self._ctx.CLAMP_TO_BORDER,  # type: ignore
                    # type: ignore
                    filter=(self._ctx.NEAREST, self._ctx.NEAREST),
                ),
            ]
        )

        self._sprites: Dict[EmbodiedEntity, Sprite] = {}

        self._playground.add_view(self)

    @property
    def texture(self):
        """The OpenGL texture containing the map pixel data"""
        return self._fbo.color_attachments[0]

    @property
    def zoom(self):
        return self._zoom

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def center(self):
        return self._center

    @property
    def sprites(self):
        return self._sprites

    def add(self, entity):

        if isinstance(entity, InteractiveAnchored):

            if not self._draw_interactive:
                return

            if self._display_uid:
                raise ValueError(
                    "Cannot display uid of interactive, set draw_interactive to False"
                )

            sprite = entity.get_sprite(self._zoom)
            self._interactive_sprites.append(sprite)

        elif isinstance(entity, InteractiveZone):

            if not self._draw_zone:
                return

            if self._display_uid:
                raise ValueError("Cannot display uid of zones, set draw_zones to False")

            sprite = entity.get_sprite(self._zoom)
            self._zone_sprites.append(sprite)

        elif isinstance(entity, PhysicalEntity):

            sprite = entity.get_sprite(self._zoom, color_uid=self._display_uid)

            if entity.traversable:
                self._traversable_sprites.append(sprite)

            elif entity.transparent:
                if self._draw_transparent:
                    self._transparent_sprites.append(sprite)

            else:
                self._visible_sprites.append(sprite)

        else:
            raise ValueError("Not implemented")

        entity.update_sprite(self, sprite)
        self._sprites[entity] = sprite

    def remove(self, entity):

        if isinstance(entity, InteractiveAnchored) and not self._draw_interactive:
            return

        if isinstance(entity, InteractiveZone) and not self._draw_zone:
            return

        sprite = self._sprites.pop(entity)

        if isinstance(entity, InteractiveAnchored):
            self._interactive_sprites.remove(sprite)

        elif isinstance(entity, InteractiveZone):
            self._zone_sprites.remove(sprite)

        elif isinstance(entity, PhysicalEntity):

            if entity.traversable:
                self._traversable_sprites.remove(sprite)

            elif entity.transparent:
                self._transparent_sprites.remove(sprite)

            else:
                self._visible_sprites.remove(sprite)

    def update_sprites(self, force=False):

        # check that
        for entity, sprite in self._sprites.items():
            if entity.needs_sprite_update or force:
                entity.update_sprite(self, sprite)

    def update(self, force=False):

        self.update_sprites(force)

        with self._fbo.activate() as fbo:

            if self._display_uid:
                fbo.clear()
            else:
                fbo.clear(self._background)

            # Change projection to match the contents
            self._ctx.projection_2d = 0, self._width, 0, self._height

            if self._draw_transparent:
                self._transparent_sprites.draw(pixelated=True)

            if self._draw_interactive:
                self._interactive_sprites.draw(pixelated=True)

            if self._draw_zone:
                self._zone_sprites.draw(pixelated=True)

            self._visible_sprites.draw(pixelated=True)
            self._traversable_sprites.draw(pixelated=True)

    def get_np_img(self):
        img = np.frombuffer(self._fbo.read(), dtype=np.dtype("B")).reshape(
            self._height, self._width, 3
        )
        return img

    def draw(self):
        img = self.get_np_img()
        plt.imshow(img)
        plt.axis("off")
        plt.show()

    def reset(self):

        self._transparent_sprites.clear()
        self._interactive_sprites.clear()
        self._zone_sprites.clear()
        self._visible_sprites.clear()
        self._traversable_sprites.clear()



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, List, Optional

import pymunk

from ..utils.definitions import (
    DEFAULT_INTERACTION_RANGE,
    INVISIBLE_ALPHA,
    CollisionTypes,
)
from .embodied import EmbodiedEntity

if TYPE_CHECKING:
    from spg.entity.physical import PhysicalEntity


class InteractiveEntity(EmbodiedEntity, ABC):
    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        for pm_shape in self._pm_shapes:
            pm_shape.sensor = True
            # pm_shape.filter = pymunk.ShapeFilter(
            #     categories=2 ** PymunkCollisionCategories.INTERACTION.value,
            # mask=2 ** PymunkCollisionCategories.INTERACTION.value)

        # self._base_sprite.alpha = INVISIBLE_ALPHA

    def get_sprite(self, zoom: float = 1, color_uid=None):
        sprite = super().get_sprite(zoom, color_uid)
        sprite.alpha = INVISIBLE_ALPHA

        return sprite

    def update_team_filter(self):

        if not self._teams:
            return

        # categ = 2 ** PymunkCollisionCategories.INTERACTION.value
        # mask= 2 ** PymunkCollisionCategories.INTERACTION.value

        categ = 0
        # categ = 2 ** PymunkCollisionCategories.INTERACTION.value
        for team in self._teams:
            categ = categ | 2 ** self._playground.teams[team]

        mask = 0
        for team in self._playground.teams:

            mask = mask | 2 ** self._playground.teams[team]
            if team not in self._teams:
                mask = mask ^ 2 ** self._playground.teams[team]

        for pm_shape in self._pm_shapes:
            pm_shape.filter = pymunk.ShapeFilter(categories=categ, mask=mask)


class InteractiveZone(InteractiveEntity, ABC):
    def _get_pm_body(self):
        return pymunk.Body(body_type=pymunk.Body.STATIC)


class InteractiveAnchored(InteractiveZone, ABC):
    def __init__(
        self,
        anchor: Optional[PhysicalEntity] = None,
        interaction_range: float = DEFAULT_INTERACTION_RANGE,
        **kwargs,
    ):

        if anchor:
            texture = anchor.texture
            radius = anchor.radius + interaction_range
            super().__init__(texture=texture, radius=radius, **kwargs)

        else:
            super().__init__(**kwargs)

        self._anchor = None

    @property
    def pm_body(self):
        return self._pm_body

    @pm_body.setter
    def pm_body(self, pm_body):
        self._pm_body = pm_body

        for pm_shape in self._pm_shapes:
            pm_shape.body = self._pm_body

    @property
    def teams(self):
        return self._teams

    @teams.setter
    def teams(self, teams):
        self._teams = teams

    @property
    def anchor(self):
        return self._anchor

    @anchor.setter
    def anchor(self, anchor: PhysicalEntity):
        self._anchor = anchor

        for pm_shape in self._pm_shapes:
            pm_shape.body = self._anchor.pm_body

        self._pm_body = self._anchor.pm_body

    def _get_pm_body(self):
        return None

    @property
    def pm_elements(self):
        return self._pm_shapes

    @property
    def needs_sprite_update(self):
        assert self._anchor
        return self._anchor.needs_sprite_update


class Graspable(InteractiveAnchored):
    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.grasped_by: List[EmbodiedEntity] = []

    @property
    def _collision_type(self):
        return CollisionTypes.GRASPABLE



from .ball import Ball
from .chest import Chest
from .diamond import Coin, Diamond
from .disabler import CommunicatorDisabler, ControllerDisabler, Disabler, SensorDisabler
from .element import (
    PhysicalElement,
    RewardElement,
    SceneElement,
    Teleporter,
    ZoneElement,
)
from .wall import (
    ColorWall,
    TiledAlternateColorWall,
    TiledGradientColorWall,
    TiledLongColorWall,
)

__all__ = [
    "Ball",
    "Chest",
    "Diamond",
    "Coin",
    "PhysicalElement",
    "RewardElement",
    "SceneElement",
    "Teleporter",
    "ZoneElement",
    "Disabler",
    "ControllerDisabler",
    "SensorDisabler",
    "CommunicatorDisabler",
    "ColorWall",
    "TiledLongColorWall",
    "TiledGradientColorWall",
    "TiledAlternateColorWall",
]



from os import path

from setuptools import find_packages, setup

with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()

# read the contents of the README file

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="spg",
    version="0.1.7",
    description="Simulator for Reinforcement Learning",
    author="Michael Garcia Ortiz",
    author_email="michael.garcia.ortiz@gmail.com",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type="text/markdown",
)



"""
Module defining the Base Classes for Sensors.
"""

from __future__ import annotations

import math
from abc import ABC, abstractmethod
from typing import List, Optional, Union

import numpy as np

from ...entity import EmbodiedEntity
from ..device import PocketDevice

SensorValue = Union[np.ndarray, List[np.ndarray]]


SENSOR_COLOR = (23, 24, 113)


class Sensor(PocketDevice):
    """Base class Sensor, used as an Interface for all sensors.

    Attributes:
        anchor: Part or Element to which the sensor is attached.
            Sensor is attached to the center of the anchor.
        sensor_values: current values of the sensor.
        name: Name of the sensor.

    Note:
        The anchor is always invisible to the sensor.

    """

    def __init__(
        self,
        normalize: Optional[bool] = False,
        **kwargs,
    ):
        """
        Sensors are attached to an anchor.
        They can detect any visible Part of an Agent or Elements of the Playground.
        If the entity is in invisible elements, it is not detected.

        Args:
            anchor: Body Part or Scene Element on which the sensor will be attached.
            normalize: boolean. If True, sensor values are scaled between 0 and 1.
            noise_params: Dictionary of noise parameters.
                Noise is applied to the raw sensor, before normalization.
            name: name of the sensor. If not provided, a name will be set by default.

        Noise Parameters:
            type: 'gaussian', 'salt_pepper'
            mean: mean of gaussian noise (default 0)
            scale: scale (or std) of gaussian noise (default 1)
            salt_pepper_probability: probability for a pixel to be turned off or max

        """

        super().__init__(color=SENSOR_COLOR, **kwargs)

        self._values = None

        self._normalize = normalize

        self._noise = False

    def update(self):

        if self._disabled:
            self._values = self._default_value

        else:
            self._compute_raw_sensor()

            if self._noise:
                self._apply_noise()

            if self._normalize:
                self._apply_normalization()

    @abstractmethod
    def _compute_raw_sensor(self):
        ...

    @abstractmethod
    def _apply_normalization(self):
        ...

    def _apply_noise(self):
        """
        Implement this method on custom sensors to add noise to the values.
        """

    @property
    @abstractmethod
    def _default_value(self) -> np.ndarray:
        ...

    @property
    @abstractmethod
    def shape(self) -> tuple:
        """Returns the shape of the numpy array, if applicable."""

    @abstractmethod
    def draw(self):
        ...


##################
# Range Sensors
##################


class ExternalSensor(Sensor, ABC):
    def __init__(
        self,
        fov: float,
        resolution: int,
        max_range: float,  # pylint: disable=redefined-builtin
        invisible_elements: Optional[
            Union[List[EmbodiedEntity], EmbodiedEntity]
        ] = None,
        invisible_when_grasped: bool = False,
        **kwargs,
    ):
        """

        Args:
            fov: Field of view of the sensor (in degrees).
            resolution: Resolution of the sensor (depends on the sensor).
            range: maximum range of the sensor (in units of distance).
            invisible_elements: Optional list of elements invisible to the sensor.
            **kwargs:

        """

        super().__init__(**kwargs)

        self._invisible_elements: List[EmbodiedEntity]

        # Invisible elements
        if not invisible_elements:
            self._invisible_elements = []
        elif isinstance(invisible_elements, EmbodiedEntity):
            self._invisible_elements = [invisible_elements]
        else:
            self._invisible_elements = invisible_elements

        self._range = max_range
        self._fov = fov * math.pi / 180
        self._resolution = resolution

        if self._resolution < 0:
            raise ValueError("resolution must be more than 1")
        if self._fov < 0:
            raise ValueError("field of view must be more than 1")
        if self._range < 0:
            raise ValueError("range must be more than 1")

        self._temporary_invisible: List[EmbodiedEntity] = []
        self._invisible_grasped = invisible_when_grasped
        self._require_invisible_update = False

    @property
    def max_range(self):
        return self._range

    @property
    def fov(self):
        return self._fov

    @property
    def resolution(self):
        return self._resolution

    @property
    def invisible_ids(self):
        return [ent.uid for ent in self._temporary_invisible + self._invisible_elements]

    @property
    def invisible_grasped(self):
        return self._invisible_grasped

    @property
    def require_invisible_update(self):
        return self._require_invisible_update

    def add_to_temporary_invisible(self, elem):
        self._temporary_invisible.append(elem)
        self._require_invisible_update = True

    def remove_from_temporary_invisible(self, elem):
        self._temporary_invisible.remove(elem)
        self._require_invisible_update = True

    def pre_step(self):
        super().pre_step()
        self._require_invisible_update = False



# pylint: disable=too-many-public-methods

""" Contains the base class for agents.

Agent class should be inherited to create agents.
It is possible to define custom Agent with
body parts, sensors and corresponding Keyboard controllers.

Examples can be found in spg/agents/agents.py

"""
from __future__ import annotations

from typing import Dict, List

from ..entity import EmbodiedEntity, Entity
from ..utils.position import Coordinate
from .communicator import Communicator
from .controller import Command, Controller
from .part import PhysicalPart
from .sensor import ExternalSensor, Sensor

_BORDER_IMAGE = 3

Commands = Dict[Controller, Command]


class Agent(Entity):

    """
    Base class for building agents.
    Agents are composed of a base and parts which are attached to the base
    or to each other.
    Each part has actuators allowing for control of the agent.
    The base has no actuator.

    Attributes:
        name: name of the agent.
            Either provided by the user or generated using internal counter.
        base: Main part of the agent. Can be mobile or fixed.
        parts: Different parts attached to the base or to other parts.
        actuators:
        sensors:
        initial_coordinates:
    """

    _index_agent: int = 0

    def __init__(
        self,
        **kwargs,
    ):

        super().__init__(**kwargs)

        # Body parts
        self._parts: List[PhysicalPart] = []

        # Reward
        self._reward: float = 0

        self._initial_coordinates = None
        self._allow_overlapping = False

    def add(self, part: PhysicalPart):
        part.agent = self
        part.teams = self._teams

        for device in part.devices:
            device.teams = self._teams

        self._parts.append(part)

    ################
    # Properties
    ################

    @property
    def base(self):
        return self._parts[0]

    @property
    def initial_coordinates(self):
        return self._initial_coordinates

    @initial_coordinates.setter
    def initial_coordinates(self, init_coord):
        self._initial_coordinates = init_coord
        self._parts[0].initial_coordinates = init_coord

    @property
    def allow_overlapping(self):
        return self._allow_overlapping

    @allow_overlapping.setter
    def allow_overlapping(self, allow):
        self._allow_overlapping = allow
        self._parts[0].allow_overlapping = allow

    @property
    def position(self):
        return self.base.position

    @property
    def angle(self):
        return self.base.angle

    ################
    # Observations
    ################

    @property
    def observations(self):
        return {
            sens: sens.sensor_values
            for part in self._parts
            for sens in part.devices
            if isinstance(sens, Sensor)
        }

    @property
    def controllers(self):
        return [
            contr
            for part in self._parts
            for contr in part.devices
            if isinstance(contr, Controller)
        ]

    @property
    def _name_to_controller(self):
        return {contr.name: contr for contr in self.controllers}

    @property
    def communicators(self):
        return [
            comm
            for part in self._parts
            for comm in part.devices
            if isinstance(comm, Communicator)
        ]

    @property
    def sensors(self):
        return [
            sensor
            for part in self._parts
            for sensor in part.devices
            if isinstance(sensor, Sensor)
        ]

    @property
    def external_sensors(self):
        return [sensor for sensor in self.sensors if isinstance(sensor, ExternalSensor)]

    def compute_observations(self):
        for sensor in self.sensors:
            sensor.update()

    ################
    # Commands
    ################

    @property
    def parts(self):
        return self._parts

    @property
    def default_commands(self) -> Commands:
        return {controller: controller.default for controller in self.controllers}

    def receive_commands(self, commands: Commands):

        for controller, command in commands.items():
            controller = self._name_to_controller[controller]
            assert controller.agent is self
            controller.command = command

    def apply_commands(self):
        # Apply command to playground physics
        for part in self._parts:
            part.apply_commands()

    def get_random_commands(self):
        return {contr.name: contr.get_random_commands() for contr in self.controllers}

    ################
    # Rewards
    ################

    @property
    def reward(self):
        return self._reward

    @reward.setter
    def reward(self, rew):
        self._reward = rew

    #############
    # ADD PARTS AND SENSORS
    #############

    def update_team_filter(self):
        for part in self._parts:
            part.update_team_filter()

    ##############
    # CONTROL
    ##############

    def pre_step(self, **kwargs):
        """
        Reset actuators and reward to 0 before a new step of the environment.
        """

        self._reward = 0

        for part in self._parts:
            part.pre_step(**kwargs)

    def reset(self):
        for part in self._parts:
            part.reset()

    def post_step(self, **kwargs):
        for part in self._parts:
            part.post_step(**kwargs)

    ###############
    # PLAYGROUND INTERACTIONS
    ###############

    def move_to(self, coord: Coordinate, **kwargs):
        """
        After moving, the agent body is back in its original configuration.
        Default angle, etc.
        """
        self.base.move_to(coordinates=coord, move_anchors=True, **kwargs)

    def _overlaps(
        self,
        entity: EmbodiedEntity,
    ) -> bool:

        assert self._playground

        for part in self.parts:
            if self._playground.overlaps(part, entity):
                return True

        return False



from __future__ import annotations

from typing import Optional, Tuple

from pymunk import Circle

from spg.utils.sprite import get_texture_from_shape

from ..entity import InteractiveAnchored
from ..utils.definitions import CollisionTypes

DEVICE_RADIUS = 5


class Device(InteractiveAnchored):
    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(
            **kwargs,
        )
        self._disabled = False

    @property
    def teams(self):
        return self._teams

    @teams.setter
    def teams(self, teams):
        self._teams = teams

    @property
    def _collision_type(self):
        return CollisionTypes.DEVICE

    def pre_step(self):
        self._disabled = False

    def disable(self):
        self._disabled = True

    @property
    def agent(self):
        assert self._anchor
        return self._anchor.agent


class PocketDevice(Device):
    def __init__(
        self,
        color: Optional[Tuple[int, int, int]] = None,
        radius: float = DEVICE_RADIUS,
        **kwargs,
    ):

        pm_shape = Circle(None, DEVICE_RADIUS)
        texture = get_texture_from_shape(
            pm_shape, color, f"Device_{DEVICE_RADIUS}_{color}"
        )
        shape_approximation = "circle"
        radius = DEVICE_RADIUS

        super().__init__(
            texture=texture,
            shape_approximation=shape_approximation,
            radius=radius,
            **kwargs,
        )



"""
Module for Spawner
"""
from __future__ import annotations

import random
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Type

from ..element import SceneElement
from ..utils.position import Coordinate, CoordinateSampler

if TYPE_CHECKING:
    from ..playground import Playground

# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments


class Spawner:
    """
    A Spawner produces entities in a random location of the playground.
    """

    id_number = 0

    def __init__(
        self,
        element_produced: Type[SceneElement],
        production_area: CoordinateSampler,
        probability: float = 0.05,
        max_elements_in_playground: int = 10,
        production_limit: int = 30,
        entity_produced_params: Optional[Dict] = None,
        allow_overlapping: bool = True,
    ):
        """
        Spawner randomly produces a new element in a random part of the Playground.
        The element is temporary, and will disappear upon reset of the Playground.

        Args:
            element_produced: SceneElement produces.
            probability: at each step, probability of creating a new element.
            max_elements_in_playground: maximum number of element present
                in the playground at any given time.
            production_limit: total number of elements that can be produced.
            entity_produced_params: Dictionary of parameters of the element produced.
            production_area: PositionAreaSampler.
            allow_overlapping: allow overlapping for spawned entities.

        """

        self.entity_produced = element_produced
        self.location_sampler = production_area

        self.entity_produced_params = {}
        if entity_produced_params:
            self.entity_produced_params = entity_produced_params

        self.entity_produced_params["temporary"] = True
        self.probability = probability
        self.limit = max_elements_in_playground
        self.total_limit = production_limit
        self.total_produced = 0
        self.produced_entities: List[SceneElement] = []
        self.allow_overlapping = allow_overlapping

        # Internal counter to assign identity number to each entity
        self.name = "spawner_" + str(Spawner.id_number)
        Spawner.id_number += 1

        self._playground: Optional[Playground] = None

    @property
    def playground(self):
        return self._playground

    @playground.setter
    def playground(self, pg):
        self._playground = pg

    @property
    def in_playground(self):
        return bool(self._playground)

    def _can_produce(self) -> bool:
        """
        Tests if the spawner can produce a new element.
        Performs random choice and checks that it is not beyond production limit.

        Returns:
            True if it can produce a new element

        """

        return (
            len(self.produced_entities) < self.limit
            and self.total_produced < self.total_limit
            and random.random() < self.probability
        )

    def produce(
        self,
    ) -> Optional[Tuple[SceneElement, Coordinate]]:
        """

        Returns: A list of SceneElement or Agent, initial position

        """

        if self._can_produce():
            obj = self.entity_produced(**self.entity_produced_params)

            self.total_produced += 1
            self.produced_entities.append(obj)

            initial_position = self.location_sampler.sample()

            return (obj, initial_position)

        return None

    def reset(self):
        """
        Reset the spawner by resetting the total count of elements produced.
        """

        self.produced_entities = []
        self.total_produced = 0



import numpy as np
import pymunk
from arcade import Texture
from PIL import Image
from skimage.draw import disk, polygon


def get_texture_from_shape(pm_shape, color, name_texture):

    color_rgba = list(color) + [255]

    if isinstance(pm_shape, pymunk.Segment):
        radius = int(pm_shape.radius)
        length = int((pm_shape.a - pm_shape.b).length)
        img = np.zeros((radius, length, 4))
        img[:, :] = color_rgba

    elif isinstance(pm_shape, pymunk.Circle):
        radius = int(pm_shape.radius)

        img = np.zeros((2 * radius + 1, 2 * radius + 1, 4))
        rr, cc = disk((radius, radius), radius)
        img[rr, cc] = color_rgba

    elif isinstance(pm_shape, pymunk.Poly):
        vertices = pm_shape.get_vertices()

        top = max(vert[0] for vert in vertices)
        bottom = min(vert[0] for vert in vertices)
        left = min(vert[1] for vert in vertices)
        right = max(vert[1] for vert in vertices)

        w = int(right - left)
        h = int(top - bottom)

        center = int(h / 2), int(w / 2)

        img = np.zeros((h, w, 4))
        r = [y + center[0] for x, y in vertices]
        c = [x + center[1] for x, y in vertices]

        rr, cc = polygon(r, c, (h, w))

        img[rr, cc] = color_rgba

    else:
        raise ValueError

    PIL_image = Image.fromarray(np.uint8(img)).convert("RGBA")

    texture = Texture(
        name=name_texture,
        image=PIL_image,
        hit_box_algorithm="Detailed",
        hit_box_detail=1,
    )

    return texture



from ..utils.definitions import CollisionTypes
from .element import PhysicalElement, RewardElement


class Diamond(PhysicalElement, RewardElement):
    def __init__(self, chest):

        super().__init__(
            mass=10,
            filename=":spg:platformer/items/diamond_green.png",
            radius=10,
        )

        self.chest = chest

    def _set_pm_collision_type(self):
        for pm_shape in self._pm_shapes:
            pm_shape.collision_type = CollisionTypes.GEM

    @property
    def _base_reward(self) -> float:
        return 10


class Coin(PhysicalElement, RewardElement):
    def __init__(self, chest, color=None):

        super().__init__(
            mass=10,
            filename=":spg:spg/coin.png",
            radius=10,
            color=color,
        )

        self.chest = chest

    def _set_pm_collision_type(self):
        for pm_shape in self._pm_shapes:
            pm_shape.collision_type = CollisionTypes.GEM

    @property
    def _base_reward(self) -> float:
        return 10
